SystemOrganization addCategory: #'Croquet-Kernel-Objects'!
SystemOrganization addCategory: #'Croquet-Kernel-Support'!
SystemOrganization addCategory: #'Croquet-Kernel-Messages'!
SystemOrganization addCategory: #'Croquet-Kernel-Tests'!

!ByteArray methodsFor: '*Croquet-Kernel-printing' stamp: 'jrd 6/10/2009 22:08'!
hexLowerCase
	| stream |
	stream := '' writeStream.
	self do: [ :each |
		stream
			nextPut: ('0123456789abcdef' at: each // 16 + 1);
			nextPut: ('0123456789abcdef' at: each \\ 16 + 1)].
	^ stream contents! !

ByteArray variableByteSubclass: #TObjectID
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Objects'!

!TObjectID commentStamp: '<historical>' prior: 0!
I represent 128 bit object IDs.!

!TObjectID class methodsFor: 'instance creation' stamp: 'bgf 5/16/2006 21:45'!
for: anObject

	"Generate a TObjectID.  
	The result depends only on the active Island's random stream.
	If logging is enabled, anObject will be logged with the new ID.
	The argument anObject plays no role other than being logged."

	| result |
	true ifTrue: [
		^ self new: 16. 	"fast path - no logging"
	].

	result := self new:16.

	"If the fast path is disabled, the following code prints
	 a transcript of every Croquet TObjectID as it is generated.
	 Slow, but helpful for pinpointing problems with replication divergence.
	 We transcript only the ID's claimed in Croquet islands...
	 to cut noise from the controller/router messaging system."

	(Processor activeIsland isKindOf: TIsland) ifTrue: [
		Transcript cr; 
			show: result asString, ' for ', anObject, ' in ', (Processor activeIsland name).
	].
	^result
! !

!TObjectID class methodsFor: 'islands' stamp: 'ar 9/15/2005 17:16'!
howToPassAsArgument
	^#passByClone:! !

!TObjectID class methodsFor: 'instance creation' stamp: 'bgf 5/15/2006 14:49'!
new
	"Generate a TObjectID.  Deprecated; #for:anObject is preferred."

	^self for: 'Unattributed'! !

!TObjectID class methodsFor: 'instance creation' stamp: 'ar 9/17/2005 17:59'!
nilObjectID
	^self basicNew: 16! !

!TObjectID class methodsFor: 'instance creation' stamp: 'ar 1/5/2006 13:07'!
readHexFrom: aStream
	^self nilObjectID readHexFrom: (aStream isString 
		ifTrue:[aStream readStream] 
		ifFalse:[aStream])! !

!TObjectID class methodsFor: 'instance creation' stamp: 'ar 1/5/2006 13:21'!
readUIDFrom: aStream
	^self nilObjectID readUIDFrom: (aStream isString 
		ifTrue:[aStream readStream] 
		ifFalse:[aStream])! !

!TObjectID methodsFor: 'converting' stamp: 'ar 9/15/2005 02:50'!
asByteArray
	^(ByteArray new: self size)
		replaceFrom: 1 to: self size with: self startingAt: 1;
		yourself! !

!TObjectID methodsFor: 'converting' stamp: 'ar 10/3/2005 15:23'!
asInteger
	| large |
	large := LargePositiveInteger new: 16.
	1 to: 16 do:[:i| large digitAt: i put: (self at: i)].
	^large normalize! !

!TObjectID methodsFor: 'converting' stamp: 'ar 9/15/2005 03:27'!
asString
	^self printString! !

!TObjectID methodsFor: 'converting' stamp: 'ar 1/5/2006 13:03'!
asSymbol
	^Symbol intern: self hex! !

!TObjectID methodsFor: 'printing' stamp: 'ar 1/5/2006 13:20'!
asUIDString
	^String streamContents:[:s| self printUIDOn: s].! !

!TObjectID methodsFor: 'accessing' stamp: 'ar 1/4/2006 12:52'!
bitXor: mask
	"XOR the receiver and mask"
	| result |
	result := self class nilObjectID.
	1 to: self size do:[:i|
		result at: i put: ((self at: i) bitXor: (mask at: i)).
	].
	^result! !

!TObjectID methodsFor: 'printing' stamp: 'ar 1/5/2006 13:03'!
hex
	"Answer a hexa decimal representation of the receiver"
	| string v index map |
	map := '0123456789abcdef'.
	string := String new: self size * 2. "hex"
	index := 0.
	1 to: self size do:[:i| 
		v := self at: i.
		string at: (index := index + 1) put: (map at: (v bitShift: -4) + 1). 
		string at: (index := index + 1) put: (map at: (v bitAnd: 15) + 1).
	].
	^string! !

!TObjectID methodsFor: 'initialize' stamp: 'ar 1/28/2006 21:48'!
initialize
	"Initialize the receiver"
	Processor activeIsland randomStream next: self size into: self startingAt: 1.
! !

!TObjectID methodsFor: 'testing' stamp: 'ar 10/3/2005 15:43'!
isZero
	1 to: self size do:[:i| (self at: i) = 0 ifFalse:[^false]].
	^true! !

!TObjectID methodsFor: 'printing' stamp: 'ar 1/5/2006 13:14'!
printHexAt: index on: aStream
	| map v |
	map := '0123456789abcdef'.
	v := self at: index.
	aStream nextPut: (map at: (v bitShift: -4) + 1). 
	aStream nextPut: (map at: (v bitAnd: 15) + 1).
! !

!TObjectID methodsFor: 'printing' stamp: 'ar 1/5/2006 13:46'!
printOn: aStream
	aStream nextPutAll: 'OID:['.
	aStream nextPutAll: self hex.
	aStream nextPut: $].! !

!TObjectID methodsFor: 'printing' stamp: 'ar 1/5/2006 13:19'!
printUIDOn: aStream
	1 to: 4 do:[:i| self printHexAt: i on: aStream].
	aStream nextPut: $-.
	5 to: 6 do:[:i| self printHexAt: i on: aStream].
	aStream nextPut: $-.
	7 to: 8 do:[:i| self printHexAt: i on: aStream].
	aStream nextPut: $-.
	9 to: 10 do:[:i| self printHexAt: i on: aStream].
	aStream nextPut: $-.
	11 to: 16 do:[:i| self printHexAt: i on: aStream].
! !

!TObjectID methodsFor: 'converting' stamp: 'ar 12/23/2005 15:08'!
readFrom: aStream
	| map v ch value |
	map := '0123456789abcdef'.
	(aStream peekFor: $( ) ifFalse:[^self error: 'Not an object id'].
	1 to: self size do:[:i|
		ch := aStream next.
		v := (map indexOf: ch) - 1.
		(v between: 0 and: 15) ifFalse:[^self error: 'Not an object id'].
		value := v bitShift: 4.
		ch := aStream next.
		v := (map indexOf: ch) - 1.
		(v between: 0 and: 15) ifFalse:[^self error: 'Not an object id'].
		value := value + v.
		self at: i put: value.
	].
	(aStream peekFor: $) ) ifFalse:[^self error: 'Not an object id'].! !

!TObjectID methodsFor: 'printing' stamp: 'ar 1/5/2006 13:17'!
readHexAt: index from: aStream
	| map v ch value |
	map := '0123456789abcdef'.
	ch := aStream next.
	v := (map indexOf: ch) - 1.
	(v between: 0 and: 15) ifFalse:[^self error: 'Not an object id'].
	value := v bitShift: 4.
	ch := aStream next.
	v := (map indexOf: ch) - 1.
	(v between: 0 and: 15) ifFalse:[^self error: 'Not an object id'].
	value := value + v.
	self at: index put: value.! !

!TObjectID methodsFor: 'printing' stamp: 'ar 1/5/2006 13:03'!
readHexFrom: aStream
	| map v ch value |
	map := '0123456789abcdef'.
	1 to: self size do:[:i|
		ch := aStream next.
		v := (map indexOf: ch) - 1.
		(v between: 0 and: 15) ifFalse:[^self error: 'Not an object id'].
		value := v bitShift: 4.
		ch := aStream next.
		v := (map indexOf: ch) - 1.
		(v between: 0 and: 15) ifFalse:[^self error: 'Not an object id'].
		value := value + v.
		self at: i put: value.
	].! !

!TObjectID methodsFor: 'printing' stamp: 'ar 1/5/2006 13:21'!
readUIDFrom: aStream
	1 to: 4 do:[:i| self readHexAt: i from: aStream].
	(aStream peekFor: $-) ifFalse:[^self error: 'Not an object id'].
	5 to: 6 do:[:i| self readHexAt: i from: aStream].
	(aStream peekFor: $-) ifFalse:[^self error: 'Not an object id'].
	7 to: 8 do:[:i| self readHexAt: i from: aStream].
	(aStream peekFor: $-) ifFalse:[^self error: 'Not an object id'].
	9 to: 10 do:[:i| self readHexAt: i from: aStream].
	(aStream peekFor: $-) ifFalse:[^self error: 'Not an object id'].
	11 to: 16 do:[:i| self readHexAt: i from: aStream].
! !

!TObjectID methodsFor: 'copying' stamp: 'sn 2/22/2018 17:25'!
shallowCopy
	^super shallowCopy! !

!TObjectID methodsFor: 'accessing' stamp: 'ar 9/15/2005 19:32'!
species
	^ByteArray! !

ByteArray variableByteSubclass: #TObjectProxy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TObjectProxy commentStamp: '<historical>' prior: 0!
I am a serializable name for an object reference.!

TObjectProxy variableByteSubclass: #TObjectIslandProxy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TObjectIslandProxy methodsFor: 'as yet unclassified' stamp: 'jrd 5/28/2009 01:08'!
comeFullyUpOnReload: refStream
	"Reconstitute me"
self halt.
	refStream class == TMessageEncoder
		ifTrue:[
			^refStream island
				lookup: self
				ifAbsent:[refStream island errorNoSuchObject: self].
		]	.! !

!TObjectProxy methodsFor: 'accessing' stamp: 'das 1/19/2006 17:00'!
comeFullyUpOnReload: refStream
	"Reconstitute me"
	refStream class == TMessageEncoder ifTrue:[
		^refStream island lookup: self ifAbsent:[refStream island errorNoSuchObject: self].
	].! !

!TObjectProxy methodsFor: 'accessing' stamp: 'ar 9/15/2005 23:54'!
species
	^ByteArray! !

IslandWriter subclass: #TSnapshotWriter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TSnapshotWriter commentStamp: '<historical>' prior: 0!
I am an experimental IslandWriter used for replication.!

!TSnapshotWriter class methodsFor: 'as yet unclassified' stamp: 'ar 9/16/2005 00:05'!
archiveVersionID
	^UUID fromString: 'b51bf9e4-9c4a-7845-bef8-1dea48449177'! !

!TSnapshotWriter class methodsFor: 'as yet unclassified' stamp: 'ar 9/16/2005 00:05'!
objectVersionID
	^UUID fromString: '66038dec-4b5e-4b40-a1a6-46fc9cdf6bb1'! !

!TSnapshotWriter methodsFor: 'export' stamp: 'ar 9/16/2005 00:04'!
exportFarRef: farRef
	^self exportNil: farRef! !

!TSnapshotWriter methodsFor: 'accessing' stamp: 'ar 9/17/2005 20:37'!
lockRefsDuring: aBlock
	| farRefLock |
	farRefLock := TFarRef allFarRefs asArray.
	farRefLock. "fake usage"
	^super lockRefsDuring: aBlock! !

!TSnapshotWriter methodsFor: 'accessing' stamp: 'ar 9/17/2005 19:36'!
rootsOf: anIsland

	^anIsland send:[:obj| obj roots].! !

IslandReader subclass: #TSnapshotReader
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TSnapshotReader commentStamp: '<historical>' prior: 0!
I am an experimental IslandReader used for replication.!

!TSnapshotReader class methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:36'!
archiveVersionID
	^TSnapshotWriter archiveVersionID! !

!TSnapshotReader class methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:36'!
objectVersionID
	^TSnapshotWriter objectVersionID! !

!UndefinedObject methodsFor: '*Croquet-Kernel' stamp: 'jrd 12/1/2009 16:15'!
markSelfAndConnectedControllers

	! !

!UndefinedObject methodsFor: '*Croquet-Kernel' stamp: 'jrd 1/7/2010 20:08'!
withGcDataDo: block

	^block value: nil! !

Heap subclass: #TMessageQueue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TMessageQueue commentStamp: '<historical>' prior: 0!
I am a message queue sorting elements primarily by their time values and secondarily by their sequence number.!

!TMessageQueue methodsFor: 'as yet unclassified' stamp: 'sn 2/23/2018 22:20'!
downHeapSingle: anIndex
	"This version is optimized for the case when only one element in the receiver can be at a wrong position. It avoids one comparison at each node when travelling down the heap and checks the heap upwards after the element is at a bottom position. Since the probability for being at the bottom of the heap is much larger than for being somewhere in the middle this version should be faster."
	| value k n j |
	anIndex = 0 ifTrue:[^self].
	n := tally bitShift: -1.
	k := anIndex.
	value := array at: anIndex.
	[k <= n] whileTrue:[
		j := k + k.
		"use max(j,j+1)"
		(j < tally and:[self sorts: (array at: j+1) before: (array at: j)])
				ifTrue:[	j := j + 1].
		array at: k put: (array at: j).
		indexUpdateBlock ifNotNil: [ indexUpdateBlock value: value value: k ].
		"and try again with j"
		k := j].
	array at: k put: value.
	indexUpdateBlock ifNotNil: [ indexUpdateBlock value: value value: k ].
	self upHeap: k! !

!TMessageQueue methodsFor: 'comparing' stamp: 'ar 9/17/2005 16:10'!
sorts: msgA before: msgB
	msgA time = msgB time 
		ifFalse:[^msgA time < msgB time].
	msgA sequenceNumber = msgB sequenceNumber
		ifFalse:[^msgA sequenceNumber < msgB sequenceNumber].
	^self error: 'These messages run at the same time'! !

!TMessageQueue methodsFor: 'as yet unclassified' stamp: 'sn 2/23/2018 22:17'!
upHeap: anIndex
	"Check the heap upwards for correctness starting at anIndex.
	 Everything below anIndex is ok."
	| value k kDiv2 tmp |
	anIndex = 0 ifTrue:[^self].
	k := anIndex.
	value := array at: anIndex.
	[ (k > 1) and:[self sorts: value before: (tmp := array at: (kDiv2 := k bitShift: -1))] ] 
		whileTrue:[
			array at: k put: tmp.
			indexUpdateBlock ifNotNil: [ indexUpdateBlock value: value value: k ].
			k := kDiv2].
	array at: k put: value.
	indexUpdateBlock ifNotNil: [ indexUpdateBlock value: value value: k ]! !

Dictionary subclass: #CobaltObjectPropertiesDictionary
	instanceVariableNames: 'objectScale'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!CobaltObjectPropertiesDictionary methodsFor: 'accessing' stamp: 'eb 2/4/2009 13:47'!
at: key 
	"Override Dictionary's at: to answer the value associated with the key but just return nil if the key is absent. "

	^ self at: key ifAbsent: nil! !

MessageSend subclass: #TMessageSend
	instanceVariableNames: 'time msgId sid'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Objects'!

!TMessageSend commentStamp: '<historical>' prior: 0!
I am a Croquet message with associated time value, message id, and sequence number.!

!TMessageSend methodsFor: 'utilities' stamp: 'ar 3/23/2006 14:39'!
carrierProcess
	"Answer the carrier process associated with this message"
	^nil! !

!TMessageSend methodsFor: 'accessing' stamp: 'ar 11/8/2005 17:04'!
id
	^msgId! !

!TMessageSend methodsFor: 'accessing' stamp: 'ar 11/8/2005 17:17'!
id: aTObjectID
	msgId := aTObjectID! !

!TMessageSend methodsFor: 'utilities' stamp: 'ar 3/23/2006 14:39'!
scriptStopped: carrierScript
	"Notify the receiver that the carrierScript just stopped executing this message"! !

!TMessageSend methodsFor: 'accessing' stamp: 'ar 9/17/2005 16:11'!
sequenceNumber
	^sid! !

!TMessageSend methodsFor: 'accessing' stamp: 'ar 9/17/2005 16:11'!
sequenceNumber: aNumber
	sid := aNumber! !

!TMessageSend methodsFor: 'accessing' stamp: 'ar 9/17/2005 16:08'!
time
	^time! !

!TMessageSend methodsFor: 'accessing' stamp: 'ar 9/17/2005 16:09'!
time: aTime
	time := aTime! !

ReferenceStream subclass: #TMessageEncoder
	instanceVariableNames: 'island'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Messages'!

!TMessageEncoder commentStamp: '<historical>' prior: 0!
I encode and decode stuff for TMessageData.!

!TMessageEncoder class methodsFor: 'instance creation' stamp: 'ar 9/15/2005 23:57'!
decode: aString in: island
	^(self basicNew setStream: (ReadStream on: aString)) 
		island: island;
		next! !

!TMessageEncoder class methodsFor: 'instance creation' stamp: 'das 9/6/2005 17:14'!
new
	^self basicNew setStream: (ReadWriteStream on: (ByteArray new: 100)).! !

!TMessageEncoder methodsFor: 'accessing' stamp: 'ar 9/15/2005 23:57'!
island
	^island! !

!TMessageEncoder methodsFor: 'accessing' stamp: 'ar 9/15/2005 23:57'!
island: anIsland
	island := anIsland.! !

!TMessageEncoder methodsFor: 'encoding' stamp: 'ar 9/15/2005 23:58'!
nextPut: obj
	"@@@@ FIXME: This needs to follow the rules in IslandArgumentCopier to the letter. 
	Right now it doesn't (or where it does it is by chance) @@@@"
	| ref |
	(obj class howToPassAsArgument) == #passByProxy: ifTrue:[
		ref := obj isFarRef ifTrue:[obj] ifFalse:[obj island asFarRef: obj].
		^super nextPut: ref
	].
	^super nextPut: obj! !

!TMessageEncoder methodsFor: 'encoding' stamp: 'jrd 8/19/2009 20:51'!
sameIslandAs: tFarRef

	^island sameIslandAs: tFarRef! !

!TMessageEncoder methodsFor: 'private' stamp: 'das 9/6/2005 18:13'!
setStream: aStream
	"PRIVATE -- Initialization method."
	aStream binary.
	basePos := aStream position.	"Remember where we start.  Earlier part of file contains a class or method file-in.  Allow that to be edited.  We don't deal in absolute file locations."
	byteStream := aStream.
	references := IdentityDictionary new: 17.
	objects := IdentityDictionary new: 17.
	fwdRefEnds := IdentityDictionary new.
	skipping := IdentitySet new.
	insideASegment := false.
	blockers ifNil: [blockers := IdentityDictionary new].	"keep blockers we just passed in"
! !

!BlockClosure methodsFor: '*Croquet-Kernel' stamp: 'sn 2/23/2018 21:43'!
croquetForkAt: priority 
	"Create and schedule a Process running the code in the receiver at the given priority. Answer the newly created process."

	^ self croquetNewProcess
		priority: priority;
		resume! !

!BlockClosure methodsFor: '*Croquet-Kernel' stamp: 'sn 2/23/2018 21:42'!
croquetNewProcess
	"Answer a Process running the code in the receiver. The process is not 
	scheduled."
	<primitive: 19> "Simulation guard"
	
^Process
		forContext: 
			[self value.
			Processor terminateActive] asContext
		priority: Processor activePriority! !

ProtoObject subclass: #TMessageMaker
	instanceVariableNames: 'myTarget'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Messages'!

!TMessageMaker commentStamp: '<historical>' prior: 0!
I convert messages into future messages.!

TMessageMaker subclass: #TFutureMaker
	instanceVariableNames: 'deltaMSecs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Messages'!

!TFutureMaker commentStamp: '<historical>' prior: 0!
I transform messages into future messages.!

!TFutureMaker methodsFor: 'accessing' stamp: 'sn 2/26/2018 16:32'!
doesNotUnderstand: aMessage
	"Package up the message and send it"
	^myTarget futureSend: aMessage selector at: deltaMSecs args: aMessage arguments! !

!TFutureMaker methodsFor: 'private' stamp: 'ar 11/3/2005 11:58'!
setDeltaMSecs: delta target: futureTarget
	deltaMSecs := delta.
	myTarget := futureTarget.! !

!TMessageMaker methodsFor: 'comparing' stamp: 'das 9/7/2005 10:10'!
= anObject
	^self == anObject! !

!TMessageMaker methodsFor: 'accessing' stamp: 'das 9/7/2005 10:08'!
basicAt: index 
	"Primitive. Assumes receiver is indexable. Answer the value of an 
	indexable element in the receiver. Fail if the argument index is not an 
	Integer or is out of bounds. Essential. Do not override in a subclass. See 
	Object documentation whatIsAPrimitive."

	<primitive: 60>
	index isInteger ifTrue: [self errorSubscriptBounds: index].
	index isNumber
		ifTrue: [^self basicAt: index asInteger]
		ifFalse: [self errorNonIntegerIndex]! !

!TMessageMaker methodsFor: 'accessing' stamp: 'das 9/7/2005 10:09'!
basicAt: index put: value 
	"Primitive. Assumes receiver is indexable. Store the second argument 
	value in the indexable element of the receiver indicated by index. Fail 
	if the index is not an Integer or is out of bounds. Or fail if the value is 
	not of the right type for this kind of collection. Answer the value that 
	was stored. Essential. Do not override in a subclass. See Object 
	documentation whatIsAPrimitive."

	<primitive: 61>
	index isInteger
		ifTrue: [(index >= 1 and: [index <= self size])
					ifTrue: [self errorImproperStore]
					ifFalse: [self errorSubscriptBounds: index]].
	index isNumber
		ifTrue: [^self basicAt: index asInteger put: value]
		ifFalse: [self errorNonIntegerIndex]! !

!TMessageMaker methodsFor: 'accessing' stamp: 'das 9/7/2005 10:10'!
basicSize
	"Primitive. Answer the number of indexable variables in the receiver. 
	This value is the same as the largest legal subscript. Essential. Do not 
	override in any subclass. See Object documentation whatIsAPrimitive."

	<primitive: 62>
	"The number of indexable fields of fixed-length objects is 0"
	^0! !

!TMessageMaker methodsFor: 'debug' stamp: 'das 9/7/2005 10:09'!
defaultLabelForInspector
	"Answer the default label to be used for an Inspector window on the receiver."
	^self class name! !

!TMessageMaker methodsFor: 'accessing' stamp: 'ar 9/16/2005 14:48'!
doesNotUnderstand: aMessage
	"Package up the message and send it"
	^myTarget syncSend: aMessage.! !

!TMessageMaker methodsFor: 'comparing' stamp: 'das 9/7/2005 10:09'!
hash
	^self identityHash! !

!TMessageMaker methodsFor: 'debug' stamp: 'das 9/7/2005 10:09'!
inspectorClass
	"Answer the class of the inspector to be used on the receiver.  Called by inspect; 
	use basicInspect to get a normal (less useful) type of inspector."

	^ Inspector! !

!TMessageMaker methodsFor: 'accessing' stamp: 'das 9/7/2005 10:09'!
instVarAt: index 
	"Primitive. Answer a fixed variable in an object. The numbering of the 
	variables corresponds to the named instance variables. Fail if the index 
	is not an Integer or is not the index of a fixed variable. Essential. See 
	Object documentation whatIsAPrimitive."

	<primitive: 73>
	"Access beyond fixed variables."
	^self basicAt: index - self class instSize		! !

!TMessageMaker methodsFor: 'accessing' stamp: 'das 9/7/2005 10:09'!
instVarAt: anInteger put: anObject 
	"Primitive. Store a value into a fixed variable in the receiver. The 
	numbering of the variables corresponds to the named instance variables. 
	Fail if the index is not an Integer or is not the index of a fixed variable. 
	Answer the value stored as the result. Using this message violates the 
	principle that each object has sovereign control over the storing of 
	values into its instance variables. Essential. See Object documentation 
	whatIsAPrimitive."

	<primitive: 74>
	"Access beyond fixed fields"
	^self basicAt: anInteger - self class instSize put: anObject! !

!TMessageMaker methodsFor: 'accessing' stamp: 'das 9/7/2005 10:09'!
isKindOf: aClass 
	"--- needed for debugging ---"
	self class == aClass
		ifTrue: [^true]
		ifFalse: [^self class inheritsFrom: aClass]! !

!TMessageMaker methodsFor: 'accessing' stamp: 'das 9/7/2005 10:09'!
isMemberOf: aClass 
	"Answer whether the receiver is an instance of the class, aClass."
	^self class == aClass! !

!TMessageMaker methodsFor: 'as yet unclassified' stamp: 'sn 2/23/2018 13:41'!
isMorph
	"Too often used and abused. Implement it here to say shut up."
	^false! !

!TMessageMaker methodsFor: 'testing' stamp: 'das 9/7/2005 10:09'!
isText
	^false! !

!TMessageMaker methodsFor: 'printing' stamp: 'das 9/7/2005 10:10'!
printOn: aStream
	"Append to the argument, aStream, a sequence of characters that  
	identifies the receiver."
	| title |
	title := self class name.
	aStream
		nextPutAll: (title first isVowel ifTrue: ['an '] ifFalse: ['a ']);
		nextPutAll: title! !

!TMessageMaker methodsFor: 'printing' stamp: 'das 9/7/2005 10:10'!
printString
	"Answer a String whose characters are a description of the receiver. 
	If you want to print without a character limit, use fullPrintString."
	^ self printStringLimitedTo: 50000! !

!TMessageMaker methodsFor: 'printing' stamp: 'das 9/7/2005 10:10'!
printStringLimitedTo: limit
	"Answer a String whose characters are a description of the receiver.
	If you want to print without a character limit, use fullPrintString."
	| limitedString |
	limitedString := String streamContents: [:s | self printOn: s] limitedTo: limit.
	limitedString size < limit ifTrue: [^ limitedString].
	^ limitedString , '...etc...'! !

!TMessageMaker methodsFor: 'private' stamp: 'das 9/6/2005 12:29'!
setTarget: aTarget
	myTarget := aTarget.! !

!TMessageMaker methodsFor: 'private' stamp: 'ar 9/16/2005 14:49'!
size
	^myTarget syncSend: (Message selector: #size arguments: #())! !

TestCase subclass: #AllCroquetTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Tests'!

!AllCroquetTests class methodsFor: 'as yet unclassified' stamp: 'ar 4/11/2006 17:43'!
addToSuiteFromSelectors: suite
	#(	CroquetVMTests CroquetSystemTests 
		IslandTests TSessionTests TMessageRouterTests
	) do:[:clsName|
		Smalltalk at: clsName ifPresent:[:tstClass|
			(tstClass withAllSubclasses asArray sort:[:c1 :c2| c1 name <= c2 name]) do: [:each | 
				each isAbstract ifFalse: [each addToSuiteFromSelectors: suite].
			].
		].
	].
! !

TestCase subclass: #CroquetSlooooooowSystemTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Tests'!

!CroquetSlooooooowSystemTests commentStamp: '<historical>' prior: 0!
Those tests that you do NOT want to run casually.!

!CroquetSlooooooowSystemTests methodsFor: 'tests' stamp: 'ar 4/11/2006 18:37'!
testDecompileAll
	"Make sure we can decompile all methods"
	CompiledMethod allInstances do:[:cm| 
		self shouldnt:[cm methodNode] raise: Error.
	].! !

!CroquetSlooooooowSystemTests methodsFor: 'tests' stamp: 'ar 4/11/2006 18:37'!
testRetrieveAllSources
	"Make sure we can retrieve the source code of all methods"
	CompiledMethod allInstances do:[:cm| 
		self shouldnt:[cm getSourceFromFile] raise: Error.
	].! !

!CroquetSlooooooowSystemTests methodsFor: 'tests' stamp: 'ar 4/11/2006 18:38'!
testRetrieveAllTimestamps
	"Make sure we can retrieve the source code of all methods"
	CompiledMethod allInstances do:[:cm| 
		self shouldnt:[cm timeStamp notNil] raise: Error.
	].! !

TestCase subclass: #CroquetSystemTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Tests'!

!CroquetSystemTests commentStamp: '<historical>' prior: 0!
A test suite to ensure some system-wide properties for Croquet.!

!CroquetSystemTests methodsFor: 'sanity' stamp: 'ar 4/11/2006 13:01'!
isPackageCategoryCovered: catName
	"Answer true if this category is covered by an MC package"
	| map cat idx |
	map := MCWorkingCopy registry.
	cat := catName.
	[map includesKey: (MCPackage named: cat)] whileFalse:[
		idx := cat lastIndexOf: $- ifAbsent:[^false].
		cat := cat copyFrom: 1 to: idx-1.
	].
	^true! !

!CroquetSystemTests methodsFor: 'compiled methods' stamp: 'ar 4/11/2006 12:10'!
testCompiledMethodsAreNew
	"ensure all compiled methods have their MethodProperties"
	CompiledMethod allInstancesDo:[:cm|
		self assert: (cm properties isKindOf: MethodProperties).
	].! !

!CroquetSystemTests methodsFor: 'compiled methods' stamp: 'ar 4/11/2006 12:10'!
testCompiledMethodsHaveMethodClass
	"ensure all compiled methods have their methodClass"
	CompiledMethod allInstancesDo:[:cm|
		self assert: (cm methodClass notNil).
	].! !

!CroquetSystemTests methodsFor: 'compiled methods' stamp: 'ar 4/11/2006 12:10'!
testCompiledMethodsHaveSelector
	"ensure all compiled methods have their selector"
	CompiledMethod allInstancesDo:[:cm|
		self assert: (cm selector notNil).
	].! !

!CroquetSystemTests methodsFor: 'sanity' stamp: 'ar 4/11/2006 12:56'!
testNoLooseExtensionMethods
	"Ensure we have no loose extension methods hanging around"
	Smalltalk allClassesDo:[:cls|
		cls organization categories do:[:cat|
			(cat at: 1) = $* ifTrue:["extension category"
				self assert: (self isPackageCategoryCovered: cat allButFirst).
			].
		].
	].! !

!CroquetSystemTests methodsFor: 'sanity' stamp: 'ar 4/11/2006 12:55'!
testNoLooseSystemCategories
	"Ensure we have no loose system categories around"
	SystemOrganization categories do:[:cat|
		self assert: (self isPackageCategoryCovered: cat).
	].! !

TestCase subclass: #CroquetVMTests
	instanceVariableNames: 'random'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Tests'!

!CroquetVMTests commentStamp: '<historical>' prior: 0!
A test suite for Croquet VM behavior!

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/17/2006 13:30'!
copysign: x from: y
	"Answer a float with the magnitude of x and the sign of y"
	| copy word1 |
	copy := x clone.
	word1 := ((x basicAt: 1) bitAnd: 16r7FFFFFFF) bitOr: ((y asFloat basicAt: 1) bitAnd: 16r80000000).
	copy basicAt: 1 put: word1.
	^copy! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:17'!
eitherPlugin: pluginName
	"see if the plugin with the given name is present as a plugin"
	| pluginString theMethod spec result |
	"try loading the plugin"
	theMethod := self class lookupSelector: #tryNamedPrimitive.
	self assert: theMethod notNil.
	theMethod ifNil:[^self].
	spec := theMethod literalAt: 1.
	spec at: 1 put: pluginName asSymbol.
	spec at: 2 put: #''.
	spec at: 3 put: 0.
	spec at: 4 put: 0.
	theMethod flushCache.
	result := self tryNamedPrimitive.
	self assert: (result == ContextPart primitiveFailToken).

	pluginString := Smalltalk listLoadedModules
		detect:[:any| any beginsWith: pluginName,' ']
		ifNone:[nil].
	self assert: pluginString notNil.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 4/11/2006 10:02'!
expectedFailures
	Smalltalk platformName = 'unix'
		ifTrue:[^#(testJoystickTabletPlugin testMIDIPlugin)].
	^#()! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:17'!
externalPlugin: pluginName
	"see if the plugin with the given name is present as a plugin"
	| pluginString theMethod spec result |
	pluginString := Smalltalk listBuiltinModules
		detect:[:any| any beginsWith: pluginName,' ']
		ifNone:[nil].
	self assert: pluginString isNil.

	"try loading the plugin"
	theMethod := self class lookupSelector: #tryNamedPrimitive.
	self assert: theMethod notNil.
	theMethod ifNil:[^self].
	spec := theMethod literalAt: 1.
	spec at: 1 put: pluginName asSymbol.
	spec at: 2 put: #''.
	spec at: 3 put: 0.
	spec at: 4 put: 0.
	theMethod flushCache.
	result := self tryNamedPrimitive.
	self assert: (result == ContextPart primitiveFailToken).

	pluginString := Smalltalk listLoadedModules
		detect:[:any| any beginsWith: pluginName,' ']
		ifNone:[nil].
	self assert: pluginString notNil.
	pluginString ifNil:[^self].
	self assert: (pluginString endsWith: ' (e)').
! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:23'!
floatTest: aBlock
	| bytes out float result hash |
	random := Random seed: 253213.
	bytes := ByteArray new: 8.
	out := WriteStream on: ByteArray new.
	float := Float basicNew: 2.
	1 to: 10000 do:[:i|
		[1 to: 8 do:[:j| bytes at: j put: (random nextInt: 256)-1].
		float basicAt: 1 put: (bytes unsignedLongAt: 1 bigEndian: true).
		float basicAt: 2 put: (bytes unsignedLongAt: 5 bigEndian: true).
		float isNaN] whileTrue.
		result := [aBlock value: float] on: Error do:[:ex|
			"we convert all errors into NaNs to have a value for testing"
			ex return: Float nan.
		].
		out nextNumber: 4 put: (result basicAt: 1).
		out nextNumber: 4 put: (result basicAt: 2).
	].
	hash := TCryptoRandom md5HashMessage: out contents.
	result := LargePositiveInteger new: 16.
	result replaceFrom: 1 to: 16 with: hash startingAt: 1.
	^result normalize! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:17'!
internalPlugin: pluginName
	"see if the plugin with the given name is present as a plugin"
	| pluginString theMethod spec result |
	pluginString := Smalltalk listBuiltinModules
		detect:[:any| any beginsWith: pluginName,' ']
		ifNone:[nil].
	self assert: pluginString notNil.
	pluginString ifNil:[^self].

	"try loading the plugin"
	theMethod := self class lookupSelector: #tryNamedPrimitive.
	self assert: theMethod notNil.
	theMethod ifNil:[^self].
	spec := theMethod literalAt: 1.
	spec at: 1 put: pluginName asSymbol.
	spec at: 2 put: #''.
	spec at: 3 put: 0.
	spec at: 4 put: 0.

	theMethod flushCache.
	result := self tryNamedPrimitive.
	self assert: (result == ContextPart primitiveFailToken).

	pluginString := Smalltalk listLoadedModules
		detect:[:any| any beginsWith: pluginName,' ']
		ifNone:[nil].
	self assert: pluginString notNil.
	pluginString ifNil:[^self].
	self assert: (pluginString endsWith: ' (i)').
! !

!CroquetVMTests methodsFor: 'miscellaneous' stamp: 'ar 3/28/2006 11:08'!
md5Hash: arg
	| hash large |
	hash := TCryptoRandom md5HashMessage: arg.
	large := LargePositiveInteger new: 16.
	large replaceFrom: 1 to: 16 with: hash startingAt: 1.
	^large normalize! !

!CroquetVMTests methodsFor: 'socket tests' stamp: 'ikp 4/14/2006 16:26'!
quickLatencyTest
	| server client accepted buf sem1 sem2 procFinished sendTimes time baselineTimes start |
	[
		client := Socket tcpCreateIfFail: [self error: 'socket create failure'].
		server := Socket tcpCreateIfFail: [self error: 'socket create failure'].
		server listenOn: 35553 backlogSize: 8.
		client setOption: 'TCP_NODELAY' value: 1.
		client connectTo: (ByteArray withAll:#(127 0 0 1)) port: 35553.
		accepted := server waitForAcceptFor: 10.
		self deny: accepted == nil.
		buf := ByteArray new: 256.
		client sendData: buf.
		accepted waitForDataIfClosed: [self error: 'connection not working'].
		accepted discardReceivedData.
		sem1 := Semaphore new.

		sendTimes := OrderedCollection new.
		"Do this for a fixed time period, not a fixed number of runs.  
		Reason: when this is running on a platform that it doesn't work 
		well on, it runs orderes of magnitude slower."
		start := Time millisecondClockValue.
		[(Time millisecondsSince: start) < 5000] whileTrue: [
			time := [	"send date; measure roundtrip time"
						procFinished := false.
						accepted discardReceivedData.
						[sem1 signal.
						accepted waitForDataIfClosed: [self halt].
						procFinished := true] forkAt: Processor userInterruptPriority.
						sem1 wait.
						client sendData: buf.
						[procFinished] whileFalse
					] timeToRun.
			sendTimes add: time
		].

		sem2 := Semaphore new.
		baselineTimes := OrderedCollection new.
		1000 timesRepeat: [
			time := [	"measure baseline time"
						procFinished := false.
						accepted discardReceivedData.
						[sem1 signal.
						sem2 wait.
						procFinished := true] forkAt: Processor userInterruptPriority.
						sem1 wait.
						sem2 signal.
						[procFinished] whileFalse.
					] timeToRun.
			baselineTimes add: time
		].
	] ensure:[
		client ifNotNil: [client destroy].
		server ifNotNil: [server destroy].
		accepted ifNotNil: [accepted destroy].
	].
	^ {
		sendTimes average asFloat. 
		sendTimes max. 
		baselineTimes average asFloat. 
		baselineTimes max 
	}! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:16'!
testADPCMCodecPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'ADPCMCodecPlugin'].
	self eitherPlugin: 'ADPCMCodecPlugin'.! !

!CroquetVMTests methodsFor: 'miscellaneous' stamp: 'ar 3/28/2006 12:03'!
testARC4Basic
	"Ensure that ARC4 works as advertised"
	"ARC4 tests vectors from OpenSSL (crypto/rc4/rc4test.c)"
	| keys inputs outputs buffer cypher |
	keys := #(
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16rEF 16r01 16r23 16r45)
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16rEF 16r01 16r23 16r45)
	).
	inputs := #(
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 
		 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00
		 16r00 16r00 16r00 16r00)
		(16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
	).
	outputs := #(
		(16r75 16rB7 16r87 16r80 16r99 16rE0 16rC5 16r96)
		(16r74 16r94 16rC2 16rE7 16r10 16r4B 16r08 16r79)
		(16rDE 16r18 16r89 16r41 16rA3 16r37 16r5D 16r3A)
		(16rD6 16rA1 16r41 16rA7 16rEC 16r3C 16r38 16rDF
		 16rBD 16r61 16r5A 16r11 16r62 16rE1 16rC7 16rBA
		 16r36 16rB6 16r78 16r58)
		(16r66 16rA0 16r94 16r9F 16r8A 16rF7 16rD6 16r89
		 16r1F 16r7F 16r83 16r2B 16rA8 16r33 16rC0 16r0C
		 16r89 16r2E 16rBE 16r30 16r14 16r3C 16rE2 16r87
		 16r40 16r01 16r1E 16rCF)
		(16rD6 16rA1 16r41 16rA7 16rEC 16r3C 16r38 16rDF 16rBD 16r61)
	).
	1 to: keys size do:[:i|
		buffer := (inputs at: i) asByteArray.
		cypher := TARC4Cypher new initKey: (keys at: i).
		cypher process: buffer.
		self assert: buffer = (outputs at: i) asByteArray.
		cypher := TARC4Cypher new initKey: (keys at: i).
		cypher process: buffer.
		self assert: buffer = (inputs at: i) asByteArray.
	].! !

!CroquetVMTests methodsFor: 'miscellaneous' stamp: 'ar 3/28/2006 12:06'!
testARC4Full
	"Ensure that ARC4 works as advertised"
	"ARC4 tests vectors from OpenSSL (crypto/rc4/rc4test.c)"
	| keys inputs outputs buffer cypher |
	keys := #(
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16rEF 16r01 16r23 16r45)
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16rEF 16r01 16r23 16r45)
	).
	inputs := #(
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 
		 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00
		 16r00 16r00 16r00 16r00)
		(16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
	).
	outputs := #(
		(16r76 16r28 16rCB 16r4A 16rA7 16r15 16r65 16r67)
		(16r77 16rB 16r8E 16r2D 16r2E 16rBE 16rA8 16r88)
		(16r9A 16r88 16r98 16rEB 16rED 16r25 16r89 16r53)
		(16r5B 16rD5 16r48 16rA1 16r58 16r3E 16r81 16r87 
		16rFB 16rE5 16r1B 16rD1 16rCC 16rFD 16r7F 16rDA 
		16r5C 16rD5 16rA9 16rB7)
		(16r65 16r3F 16rD8 16r55 16rB4 16r2 16r76 16r78 
		16rF3 16r5A 16rC 16r70 16r89 16rC8 16r1C 16r3B 
		16rA9 16r30 16rF5 16r6D 16rF9 16rD 16r2D 16r84 
		16rC3 16r83 16r62 16r17)
		(16r5B 16rD5 16r48 16rA1 16r58 16r3E 16r81 16r87 16rFB 16rE5)
	).
	1 to: keys size do:[:i|
		buffer := (inputs at: i) asByteArray.
		cypher := TARC4Cypher new key: (keys at: i).
		cypher process: buffer.
		self assert: buffer = (outputs at: i) asByteArray.
		cypher := TARC4Cypher new key: (keys at: i).
		cypher process: buffer.
		self assert: buffer = (inputs at: i) asByteArray.
	].! !

!CroquetVMTests methodsFor: 'miscellaneous' stamp: 'ar 3/28/2006 12:05'!
testARC4Partial
	"Ensure that ARC4 works as advertised"
	"ARC4 tests vectors from OpenSSL (crypto/rc4/rc4test.c)"
	| keys inputs outputs buffer cypher |
	keys := #(
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16rEF 16r01 16r23 16r45)
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16rEF 16r01 16r23 16r45)
	).
	inputs := #(
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 
		 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00
		 16r00 16r00 16r00 16r00)
		(16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
	).
	outputs := #(
		(16r76 16r28 16rCB 16r4A 16rA7 16r15 16r65 16r67)
		(16r77 16rB 16r8E 16r2D 16r2E 16rBE 16rA8 16r88)
		(16r9A 16r88 16r98 16rEB 16rED 16r25 16r89 16r53)
		(16r5B 16rD5 16r48 16rA1 16r58 16r3E 16r81 16r87 
		16rFB 16rE5 16r1B 16rD1 16rCC 16rFD 16r7F 16rDA 
		16r5C 16rD5 16rA9 16rB7)
		(16r65 16r3F 16rD8 16r55 16rB4 16r2 16r76 16r78 
		16rF3 16r5A 16rC 16r70 16r89 16rC8 16r1C 16r3B 
		16rA9 16r30 16rF5 16r6D 16rF9 16rD 16r2D 16r84 
		16rC3 16r83 16r62 16r17)
		(16r5B 16rD5 16r48 16rA1 16r58 16r3E 16r81 16r87 16rFB 16rE5)
	).
	1 to: keys size do:[:i|
		buffer := (inputs at: i) asByteArray.
		cypher := TARC4Cypher new key: (keys at: i).
		"exercise partial processing"
		1 to: buffer size do:[:j| cypher process: buffer from: j to: j].
		self assert: buffer = (outputs at: i) asByteArray.
		cypher := TARC4Cypher new key: (keys at: i).
		"exercise partial processing"
		1 to: buffer size do:[:j| cypher process: buffer from: j to: j].
		self assert: buffer = (inputs at: i) asByteArray.
	].! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/11/2006 10:03'!
testArcCos
	| hash |
	hash := self floatTest:[:f| f arcCos].
	self assert: hash = 320603091210691421897131240956682310429! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:52'!
testArcCosH
	| hash |
	hash := self floatTest:[:f| f arcCosH].
	self assert: hash = 6724426144112251941037505276242428134! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:27'!
testArcCosHStd
	| inputA resultA inputB resultB |
	{	"test input value"	"test result plus"		"test result minus"
		{	0.0.				nil.					nil	}.
		{	1.0.				'0000000000000000'.	nil	}.
		{	2.0.				'3FF5124271980434'.	nil	}.
		{	Float infinity.	'7FF0000000000000'.	nil	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA arcCosH hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB arcCosH hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:28'!
testArcCosStd
	| inputA resultA inputB resultB |
	{	"test input value"		"test result plus"		"test result minus"
		{	0.0.				'3FF921FB54442D18'.	'3FF921FB54442D18'	}.
		{	1.0.				'0000000000000000'.	'400921FB54442D18'	}.
		{	2.0.				nil.					nil	}.
		{	Float infinity.	nil.					nil	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA arcCos hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB arcCos hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:52'!
testArcSin
	| hash |
	hash := self floatTest:[:f| f arcSin].
	self assert: hash = 27372132577303862731837100895783885417! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:52'!
testArcSinH
	| hash |
	hash := self floatTest:[:f| f arcSinH].
	self assert: hash = 255911863578190171815115260235896145802! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:33'!
testArcSinHStd
	| inputA resultA inputB resultB |
	{	"test input value"	"test result plus"		"test result minus"
		{	0.0.				'0000000000000000'.	'8000000000000000'	}.
		{	1.0.				'3FEC34366179D427'.	'BFEC34366179D427'	}.
		{	2.0.				'3FF719218313D087'.	'BFF719218313D087'	}.
		{	Float infinity.	'7FF0000000000000'.	'FFF0000000000000'	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA arcSinH hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB arcSinH hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:34'!
testArcSinStd
	| inputA resultA inputB resultB |
	{	"test input value"	"test result plus"		"test result minus"
		{	0.0.				'0000000000000000'.	'8000000000000000'	}.
		{	1.0.				'3FF921FB54442D18'.	'BFF921FB54442D18'	}.
		{	2.0.				nil.					nil	}.
		{	Float infinity.	nil.					nil	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA arcSin hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB arcSin hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].
! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:52'!
testArcTan
	| hash |
	hash := self floatTest:[:f| f arcTan].
	self assert: hash = 17311773710959114634056077345168823659! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:53'!
testArcTan2
	| hash |
	hash := self floatTest:[:f| f arcTan: f].
	self assert: hash = 287068347279655848752274030373495709564! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:53'!
testArcTanH
	| hash |
	hash := self floatTest:[:f| f arcTanH].
	self assert: hash = 295711907369004359459882231908879164929! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:35'!
testArcTanHStd
	| inputA resultA inputB resultB |
	{	"test input value"	"test result plus"		"test result minus"
		{	0.0.				'0000000000000000'.	'8000000000000000'	}.
		{	1.0.				'7FF0000000000000'.	'FFF0000000000000'	}.
		{	2.0.				nil.					nil	}.
		{	Float infinity.	nil.					nil	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA arcTanH hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB arcTanH hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:35'!
testArcTanStd
	| inputA resultA inputB resultB |
	{	"test input value"	"test result plus"		"test result minus"
		{	0.0.				'0000000000000000'.	'8000000000000000'	}.
		{	1.0.				'3FE921FB54442D18'.	'BFE921FB54442D18'	}.
		{	2.0.				'3FF1B6E192EBBE44'.	'BFF1B6E192EBBE44'	}.
		{	Float infinity.	'3FF921FB54442D18'.	'BFF921FB54442D18'	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA arcTan hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB arcTan hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:15'!
testAsynchFilePlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'AsynchFilePlugin'].
	self eitherPlugin: 'AsynchFilePlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:15'!
testB3DAcceleratorPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'B3DAcceleratorPlugin'].
	self eitherPlugin: 'B3DAcceleratorPlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:15'!
testBMPReadWriterPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'BMPReadWriterPlugin'].
	self eitherPlugin: 'BMPReadWriterPlugin'.! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:53'!
testCos
	| hash |
	hash := self floatTest:[:f| f cos].
	self assert: hash = 110207739557966732640546618158077332978! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:53'!
testCosH
	| hash |
	hash := self floatTest:[:f| f cosH].
	self assert: hash = 139309299067563830037108641802292492276! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:35'!
testCosHStd
	| inputA resultA inputB resultB |
	{	"test input value"		"test result plus"		"test result minus"
		{	0.0.				'3FF0000000000000'.	'3FF0000000000000'	}.
		{	1.0.				'3FF8B07551D9F551'.	'3FF8B07551D9F551'	}.
		{	2.0.				'400E18FA0DF2D9BC'.	'400E18FA0DF2D9BC'	}.
		{	Float infinity.	'7FF0000000000000'.	'7FF0000000000000'	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA cosH hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB cosH hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].
! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:36'!
testCosStd
	| inputA resultA inputB resultB |
	{	"test input value"	"test result plus"		"test result minus"
		{	0.0.				'3FF0000000000000'.	'3FF0000000000000'	}.
		{	1.0.				'3FE14A280FB5068C'.	'3FE14A280FB5068C'	}.
		{	2.0.				'BFDAA22657537205'.	'BFDAA22657537205'	}.
		{	Float infinity.	nil.					nil	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA cos hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB cos hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:15'!
testCroquetPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'CroquetPlugin'].
	self eitherPlugin: 'CroquetPlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:15'!
testDropPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'DropPlugin'].
	self eitherPlugin: 'DropPlugin'.! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:03'!
testExp
	| hash |
	hash := self floatTest:[:f| f exp].
	self assert: hash = 264681209343177480335132131244505189510! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:36'!
testExpStd
	| inputA resultA inputB resultB |
	{	"test input value"		"test result plus"		"test result minus"
		{	0.0.				'3FF0000000000000'.	'3FF0000000000000'	}.
		{	1.0.				'4005BF0A8B14576A'.	'3FD78B56362CEF38'	}.
		{	2.0.				'401D8E64B8D4DDAE'.	'3FC152AAA3BF81CC'	}.
		{	Float infinity.	'7FF0000000000000'.	'0000000000000000'	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA exp hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB exp hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:15'!
testFFTPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'FFTPlugin'].
	self eitherPlugin: 'FFTPlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:15'!
testFilePlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'FilePlugin'].
	self eitherPlugin: 'FilePlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:15'!
testFloatArrayPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'FloatArrayPlugin'].
	self eitherPlugin: 'FloatArrayPlugin'.! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:03'!
testFloatAt
	| hash flt |
	flt := FloatArray new: 1.
	hash := self floatTest:[:f| flt at: 1 put: f. flt at: 1].
	self assert: hash = 80498428122197125691266588764018905399! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:15'!
testFloatMathPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'FloatMathPlugin'].
	self eitherPlugin: 'FloatMathPlugin'.! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:03'!
testFraction
	| hash |
	hash := self floatTest:[:f| f fractionPart].
	self assert: hash = 320444785026869345695277323179170692004! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:31'!
testFractionStd
	| inputA resultA inputB resultB |
	{	"test input value"	"test result plus"		"test result minus"
		{	0.0.				'0000000000000000'.	'8000000000000000'	}.
		{	1.0.				'0000000000000000'.	'8000000000000000'	}.
		{	2.0.				'0000000000000000'.	'8000000000000000'	}.
		{	Float infinity.	'0000000000000000'.	'8000000000000000'	}.
		{	Float nan.		'0000000000000000'.	'8000000000000000'	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA fractionPart hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB fractionPart hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:15'!
testGeniePlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'GeniePlugin'].
	self eitherPlugin: 'GeniePlugin'.! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:53'!
testHypot
	| hash |
	hash := self floatTest:[:f| f hypot: f+1].
	self assert: hash = 217113721886532765853628735806816720346! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testJPEGReadWriter2Plugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'JPEGReadWriter2Plugin'].
	self eitherPlugin: 'JPEGReadWriter2Plugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testJPEGReaderPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'JPEGReaderPlugin'].
	self eitherPlugin: 'JPEGReaderPlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testJoystickTabletPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'JoystickTabletPlugin'].
	self eitherPlugin: 'JoystickTabletPlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testKlattPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'Klatt'].
	self eitherPlugin: 'Klatt'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testLargeIntegersPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'LargeIntegers'].
	self eitherPlugin: 'LargeIntegers'.! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:03'!
testLog
	| hash |
	hash := self floatTest:[:f| f abs ln].
	self assert: hash = 24389651894375564945708989023746058645! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:54'!
testLog10
	| hash |
	hash := self floatTest:[:f| f abs log].
	self assert: hash = 135564553959509933253581837789050718785! !

!CroquetVMTests methodsFor: 'miscellaneous' stamp: 'ar 3/28/2006 11:10'!
testMD5
	"Ensure that MD5 primitive works properly"
	self assert: (self md5Hash: 'a') = 16r0CC175B9C0F1B6A831C399E269772661.
	self assert: (self md5Hash: 'abc') = 16r900150983CD24FB0D6963F7D28E17F72.
	self assert: (self md5Hash: 'message digest') = 16rF96B697D7CB7938D525A2F31AAF161D0.
	self assert: (self md5Hash:
		'abcdefghijklmnopqrstuvwxyz') = 16rC3FCD3D76192E4007DFB496CCA67E13B.
	self assert: (self md5Hash:
		'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789') =
		16rD174AB98D277D9F5A5611C2C9F419D9F.
	self assert: (self md5Hash:
		'12345678901234567890123456789012345678901234567890123456789012345678901234567890') =
 		16r57EDF4A22BE3C955AC49DA2E2107B67A.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testMIDIPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'MIDIPlugin'].
	self eitherPlugin: 'MIDIPlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testMatrix2x3Plugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'Matrix2x3Plugin'].
	self eitherPlugin: 'Matrix2x3Plugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testMiscPrimitivePlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'MiscPrimitivePlugin'].
	self eitherPlugin: 'MiscPrimitivePlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testMpeg3Plugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'Mpeg3Plugin'].
	self eitherPlugin: 'Mpeg3Plugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testRePlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'RePlugin'].
	self eitherPlugin: 'RePlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testSecurityPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'SecurityPlugin'].
	self eitherPlugin: 'SecurityPlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testSerialPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'SerialPlugin'].
	self eitherPlugin: 'SerialPlugin'.! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:54'!
testSin
	| hash |
	hash := self floatTest:[:f| f sin].
	self assert: hash = 290162321010315440569513182938961037473! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:54'!
testSinH
	| hash |
	hash := self floatTest:[:f| f sinH].
	self assert: hash = 146029709156303766079448006055284064911! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:36'!
testSinHStd
	| inputA resultA inputB resultB |
	{	"test input value"	"test result plus"		"test result minus"
		{	0.0.				'0000000000000000'.	'8000000000000000'	}.
		{	1.0.				'3FF2CD9FC44EB982'.	'BFF2CD9FC44EB982'	}.
		{	2.0.				'400D03CF63B6E1A0'.	'C00D03CF63B6E1A0'	}.
		{	Float infinity.	'7FF0000000000000'.	'FFF0000000000000'	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA sinH hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB sinH hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:37'!
testSinStd
	| inputA resultA inputB resultB |
	{	"test input value"		"test result plus"		"test result minus"
		{	0.0.				'0000000000000000'.	'8000000000000000'	}.
		{	1.0.				'3FEAED548F090CEE'.	'BFEAED548F090CEE'	}.
		{	2.0.				'3FED18F6EAD1B446'.	'BFED18F6EAD1B446'	}.
		{	Float infinity.	nil.					nil	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA sin hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB sin hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'socket tests' stamp: 'ar 4/13/2006 15:02'!
testSocketBindTo
	"Test that the bind() primitive on sockets is supported"
	| socket address port |
	address := ByteArray withAll: #(127 0 0 1).
	port := 32152.
	socket := Socket newTCP.
	[
		self shouldnt:[socket bindTo: address port: port] raise: Error.
	] ensure:[socket destroy].
! !

!CroquetVMTests methodsFor: 'socket tests' stamp: 'ar 4/13/2006 15:30'!
testSocketLatency
	| results |
	results := self quickLatencyTest.
	Transcript cr; show: '-------------------- testSocketLatency -------------------'.
	Transcript crtab; show: 'Avg. send time: ', results first.
	Transcript crtab; show: 'Max send time: ', results second.
	Transcript crtab; show: 'Avg. baseline time: ', results third.
	Transcript crtab; show: 'Max baseline time: ', results fourth.
	self should: [results first < 5.0] "average time should be well under 5 msec"
! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:14'!
testSocketPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'SocketPlugin'].
	self eitherPlugin: 'SocketPlugin'.! !

!CroquetVMTests methodsFor: 'socket tests' stamp: 'bgf 5/14/2006 14:00'!
testSocketReusability
	"Tests whether 'aSocket beReusable' works properly."
	| address port udp1 send1 udp2 recv2 sendProc recvProc received |
	address := ByteArray withAll: #(255 255 255 255). "broadcast"
	port := 31259.
	[
		udp1 := Socket newUDP.
		udp1 setOption: 'SO_REUSEADDR' value: 1.
		udp1 setOption: 'SO_REUSEPORT' value: 1.
		udp1 setPort: port.
		udp1 setOption: 'SO_BROADCAST' value: 1.
		send1 := TObjectID for: 'Send'.

		udp2 := Socket newUDP.
		udp2 setOption: 'SO_REUSEADDR' value: 1.
		udp2 setOption: 'SO_REUSEPORT' value: 1.
		udp2 setPort: port.
		udp2 setOption: 'SO_BROADCAST' value: 1.
		recv2 := TObjectID for: 'Recv'.

		received := 0.
		recvProc := [
			[received < 16] whileTrue:[
				received := received + (udp2 receiveDataInto: recv2 startingAt: received + 1).
			]
		] fork.
		sendProc := [
			udp1 setPeer: address port: port.
			udp1 sendData: send1 count: 16.
		] fork.
		(Delay forMilliseconds: 200) wait.
		self should: [recvProc isTerminated].
		self should: [sendProc isTerminated].
		self should: [send1 = recv2].
	] ensure:[
		udp1 destroy.
		udp2 destroy.
	].
! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:13'!
testSoundCodecPrims
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'SoundCodecPrims'].
	self eitherPlugin: 'SoundCodecPrims'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:13'!
testSoundGenerationPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'SoundGenerationPlugin'].
	self eitherPlugin: 'SoundGenerationPlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:13'!
testSoundPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'SoundPlugin'].
	self eitherPlugin: 'SoundPlugin'.! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:03'!
testSqrt
	| hash |
	hash := self floatTest:[:f| f abs sqrt].
	self assert: hash = 112236588358122834093969606123302196127! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:13'!
testSqueakFFIPrims
	Smalltalk platformName = 'Win32'
		ifTrue:[^self externalPlugin: 'SqueakFFIPrims'].
	self eitherPlugin: 'SqueakFFIPrims'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:13'!
testStarSqueakPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'StarSqueakPlugin'].
	self eitherPlugin: 'StarSqueakPlugin'.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:13'!
testSurfacePlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'SurfacePlugin'].
	self eitherPlugin: 'SurfacePlugin'.! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/11/2006 10:03'!
testTan
	| hash |
	hash := self floatTest:[:f| f tan].
	self assert: hash = 207143885112027702205238433494882679660! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:54'!
testTanH
	| hash |
	hash := self floatTest:[:f| f tanH].
	self assert: hash = 15738508136206638425252880299326548123! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/19/2006 00:27'!
testTanHStd
	| inputA resultA inputB resultB |
	{	"test input value"	"test result plus"		"test result minus"

"ar 4/19/2006: I have disabled the test for tanh(-0.0) for the time being because it was literally the 'last bit' differing amongst the various VMs before Croquet 1.0 and we were all wiped out by the time we ran into it to the point that I felt that having a different result for tan(-0.0) (-0.0 on some and 0.0 on other platforms with -0.0 being the 'correct' answer) wasn't worth pushing it any further. At some point we'll put it back in (likely before the next release; by then we'll have recovered from this one ;-) and fix it for good."

"		{	0.0.				'0000000000000000'.	'8000000000000000'	}.		"
		{	1.0.				'3FE85EFAB514F394'.	'BFE85EFAB514F394'	}.
		{	2.0.				'3FEED9505E1BC3D4'.	'BFEED9505E1BC3D4'	}.
		{	Float infinity.	'3FF0000000000000'.	'BFF0000000000000'	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA tanH hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB tanH hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 4/18/2006 10:37'!
testTanStd
	| inputA resultA inputB resultB |
	{	"test input value"		"test result plus"		"test result minus"
		{	0.0.				'0000000000000000'.	'8000000000000000'	}.
		{	1.0.				'3FF8EB245CBEE3A6'.	'BFF8EB245CBEE3A6'	}.
		{	2.0.				'C0017AF62E0950F8'.	'40017AF62E0950F8'	}.
		{	Float infinity.	nil.					nil	}.
		{	Float nan.		nil.					nil	}.
	} do:[:spec|
		inputA := self copysign: spec first from: 1.0.
		resultA := [inputA tan hex] on: Error do:[:ex| ex return: nil].
		inputB := self copysign: spec first from: -1.0.
		resultB := [inputB tan hex] on: Error do:[:ex| ex return: nil].
		self assert: spec second = resultA.
		self assert: spec third = resultB.
	].! !

!CroquetVMTests methodsFor: 'float tests' stamp: 'ar 3/26/2006 17:03'!
testTimesTwoPower
	| hash |
	hash := self floatTest:[:f| f timesTwoPower: (random nextInt: 200) - 100].
	self assert: hash = 278837335583284459890979576373223649870.! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:13'!
testUUIDPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'UUIDPlugin'].
	self eitherPlugin: 'UUIDPlugin'.! !

!CroquetVMTests methodsFor: 'vm version' stamp: 'ar 3/26/2006 20:30'!
testVMBuildVersion
	"Ensure that the VM reports a build string (getSystemAttribute: 1006)"
	self deny: (Smalltalk getSystemAttribute: 1006) isEmptyOrNil! !

!CroquetVMTests methodsFor: 'plugins' stamp: 'ar 3/26/2006 23:13'!
testZipPlugin
	Smalltalk platformName = 'Win32'
		ifTrue:[^self internalPlugin: 'ZipPlugin'].
	self eitherPlugin: 'ZipPlugin'.! !

TestCase subclass: #TFarRefRegistryTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TFarRefRegistryTests methodsFor: 'as yet unclassified' stamp: 'jrd 12/29/2009 22:00'!
testAdding

	| reg |
	reg := TFarRefRegistry new.
	reg add: 'thingy'.
	reg add: 'bar'.
	reg inspect.
! !

!TFarRefRegistryTests methodsFor: 'as yet unclassified' stamp: 'jrd 12/29/2009 22:00'!
testCreation

	| reg |
	reg := TFarRefRegistry new.
	self assert: (reg isKindOf: TFarRefRegistry).
	self assert: (reg asArray isEmpty).
! !

Island subclass: #TIsland
	instanceVariableNames: 'controller id messageQueue time sequence nameMap mutex'
	classVariableNames: 'ShouldLog'
	poolDictionaries: ''
	category: 'Croquet-Kernel-Objects'!

!TIsland commentStamp: 'jrd 9/16/2009 21:29' prior: 0!
Current representation for a Croquet island. Currently undergoing changes.

Comment from sendMessage: on an important ongoing problem:
	"@@@@@FIXME: We have a potential problem here...sometimes this can be
	reached when the controller is nil, because a snapshot is in progress. There
	is a mutex that is in play during the snapshot, but it is in the controller....so
	we can't reach it. And even if we could, the datapath for sending the message
	is completely independent of the sync...it would just be a meaningless delay.
	Until we can come up with a better plan, we busy wait with a delay until we
	have a controller, then forward the message. We wait 100 ms as that
	seems to be close to a lower limit on the sync time."
	"This is one of those places where I'm not sure how it ever worked for more than
	5 min in a row...."
	"Possible better plans: have farrefs refer to the controller rather than the
	island, or maybe both. Have an islandID->controller lookup table someplace
	public in case the controller ivar is nil "
!

!TIsland class methodsFor: 'class initialization' stamp: 'mtf 1/22/2010 17:12'!
initialize
	"TIsland initialize"
	| seed |
	seed := ByteArray new: 128.
	TCryptoRandom gatherEntropy: seed.
	Island default random: Random new.
	Island default randomStream: TARC4Cypher new.
	Island default randomStream key: seed.
	ShouldLog := nil.! !

!TIsland methodsFor: 'event loop' stamp: 'sn 2/23/2018 21:46'!
advanceTo: nowTime
	"Advance this island's simulation to nowTime"
	| msg msgTime |

	[msg := messageQueue firstOrNil.
	msg == nil or:[msg time > nowTime]] whileFalse:[
		msgTime := msg time.
		messageQueue removeFirst.
		msgTime < time ifTrue:[^self errorPastMessage: msg].
		time := msgTime.
		self execute: msg.
	].
	time := nowTime.
! !

!TIsland methodsFor: 'accessing' stamp: 'ar 9/16/2005 14:04'!
asFarRef: anObject
	"Answer the far ref for the given (near) object"
	anObject isFarRef ifTrue:[^anObject].
	^exports at: anObject ifAbsentPut:[
		| ref |
		ref := TFarRef new.
		anObject myFarRef: ref.
		ref].! !

!TIsland methodsFor: 'accessing' stamp: 'ar 11/8/2005 18:11'!
asFarRef: anObject put: farRef
	"Answer the far ref for the given (near) object"
	anObject isFarRef ifTrue:[^anObject].
	^exports at: anObject ifAbsentPut:[
		anObject myFarRef: farRef.
		farRef].! !

!TIsland methodsFor: 'accessing' stamp: 'das 12/27/2005 16:57'!
basicNew: aClass
	^aClass basicNew! !

!TIsland methodsFor: 'accessing' stamp: 'das 12/27/2005 16:57'!
basicNew: aClass size: sz
	^aClass basicNew: sz! !

!TIsland methodsFor: 'registry' stamp: 'hrs 1/28/2007 12:01'!
compactNameMap: refId
	"Remove everything from nameMap that isn't reachable through globals.
	Guarnteed to break things if participants are doing someting."
	| ourFarRef |

	"We should probably also get rid of TForms that aren't referenced anywhere. Later. -H"

	nameMap := Dictionary new.
	globals do: [:root | root deepRegisterIntoIsland: self].
	ourFarRef := (self asFarRef: self).
	nameMap at: id put: ourFarRef.
	nameMap at: TObjectID nilObjectID put: ourFarRef.
	nameMap at: refId put: ourFarRef.
! !

!TIsland methodsFor: 'accessing' stamp: 'ar 9/8/2005 11:23'!
controller
	^controller! !

!TIsland methodsFor: 'accessing' stamp: 'ar 9/8/2005 11:23'!
controller: aController
	controller := aController.! !

!TIsland methodsFor: 'accessing' stamp: 'ar 1/30/2006 23:09'!
controllerID
	^controller senderID! !

!TIsland methodsFor: 'event loop' stamp: 'ar 2/1/2006 15:01'!
decode: aTMessage
	| msg obj |
	msg := TMessageSend new.
	obj := aTMessage receiver.
	msg receiver: (self lookup: obj ifAbsent:[
		self register: Object new name: obj. "try to trigger race condition"
		self errorNoSuchObject: obj]).
	msg selector: aTMessage selector.
	msg arguments: (TMessageEncoder decode: aTMessage arguments in: self).
	msg time: aTMessage time.
	msg id: aTMessage id.
	msg id isZero ifTrue:[msg id: nil].
	^msg! !

!TIsland methodsFor: 'event loop' stamp: 'jrd 4/2/2010 15:29'!
decodeScheduleAndExecute: aTMessage
	"Decode, schedule, and execute all message up to the time specified in aTMessage."
	| msg |
	"Shoul make the error logging much more explicit @@@@"
	self flag: #jrd . 
	msg := [ self decode: aTMessage ] on: Error do: [:e | 
		self log error: 'Invalid TMessage ' , e description.
		^ self.
	].
	self scheduleMessage: msg at: msg time.
	self advanceTo: msg time.! !

!TIsland methodsFor: 'registry' stamp: 'jrd 5/28/2009 01:23'!
errorNoSuchObject: aName

	self error: aName hex ,': No such object'.
	^nil



! !

!TIsland methodsFor: 'event loop' stamp: 'ar 9/17/2005 18:38'!
errorPastMessage: msg
	^self error: 'Past message encountered: ',msg selector! !

!TIsland methodsFor: 'event loop' stamp: 'ar 11/8/2005 18:49'!
execute: aTMessage
	| result id promised |
	id := aTMessage id.
	result := aTMessage receiver 
		perform: aTMessage selector 
		withArguments: aTMessage arguments.
	id ifNotNil:[
		"Name the result accordingly"
		promised := nameMap at: id ifAbsentPut:[TFarRef new].
		promised privateValue: result island: self.
		"Recycle farRef if possible"
		self asFarRef: result put: promised.
		self signal: id asSymbol with: result.
	].
! !

!TIsland methodsFor: 'accessing' stamp: 'das 11/30/2005 10:51'!
id
	^ id.
! !

!TIsland methodsFor: 'accessing' stamp: 'das 11/27/2005 16:53'!
id: i
	id := i.
	self register: self name: id.! !

!TIsland methodsFor: 'accessing' stamp: 'das 11/27/2005 17:14'!
id: i name: nm
	id := i.
	self register: self name: id.
	name := nm.! !

!TIsland methodsFor: 'initialize' stamp: 'jrd 6/12/2012 16:05'!
initialize

	super initialize.
	mutex := TMutex new.
	messageQueue := TMessageQueue new.
	time := 0.0.
	sequence := 0.
	nameMap := Dictionary new.
	self initializeRandomGenerators.
	"Setup a temporary ID for the Island so that we can process messages.
	Also, required for local islands."
	self id: (TObjectID for: self).
	"Register myself under the known NULL object ID for messages from the router"
	nameMap at: TObjectID nilObjectID put: (self asFarRef: self).! !

!TIsland methodsFor: 'initialize' stamp: 'jrd 6/12/2012 16:01'!
initializeRandom: seed
	random := Random seed: seed.
! !

!TIsland methodsFor: 'initialize' stamp: 'jrd 6/12/2012 16:05'!
initializeRandomGenerators

	| seed |
	self initializeRandom: Time totalSeconds.
	randomStream ifNil:["Could be initialized in superclass"
		seed := ByteArray new: 128.
		TCryptoRandom gatherEntropy: seed.
		self initializeRandomStream: seed.
	].
! !

!TIsland methodsFor: 'initialize' stamp: 'jrd 6/12/2012 16:06'!
initializeRandomSeeds: pair

	self initializeRandom: pair first.
	self initializeRandomStream: pair second.
! !

!TIsland methodsFor: 'initialize' stamp: 'jrd 6/12/2012 16:03'!
initializeRandomStream: seed

	randomStream := TARC4Cypher new.
	randomStream key: seed.

! !

!TIsland methodsFor: 'event loop' stamp: 'jrd 5/4/2009 01:20'!
logTMessage: msg

	ShouldLog
		ifNotNil: [ self log:  (time printString , ' ' , (msg receiver class name) , ' ' , (msg selector)) ].
! !

!TIsland methodsFor: 'registry' stamp: 'das 1/19/2006 16:57'!
lookup: aName
	^self lookup: aName ifAbsent:[nil]! !

!TIsland methodsFor: 'registry' stamp: 'ar 2/1/2006 17:08'!
lookup: aName ifAbsent: exceptionBlock
	| ref |
	ref := mutex critical:[nameMap at: aName ifAbsent:[nil]].
	ref ifNil:[^exceptionBlock value].
	^ref valueOn: self! !

!TIsland methodsFor: 'accessing' stamp: 'ar 9/14/2005 19:25'!
make: aClass
	| obj |
	obj := self new: aClass.
	self register: obj.
	^obj! !

!TIsland methodsFor: 'event loop' stamp: 'ar 9/17/2005 16:14'!
messageQueue
	^messageQueue! !

!TIsland methodsFor: 'event loop' stamp: 'ar 9/17/2005 16:14'!
messageQueue: aTMessageQueue
	^messageQueue := aTMessageQueue! !

!TIsland methodsFor: 'accessing' stamp: 'ar 11/3/2005 12:08'!
millisecondClockValue
	"Answer the number of milliseconds since the millisecond clock was last reset or rolled over."
	^time truncated bitAnd: 16r3FFFFFFF! !

!TIsland methodsFor: 'registry' stamp: 'das 9/6/2005 15:14'!
nameOf: anObject
	^(self asFarRef: anObject) myName! !

!TIsland methodsFor: 'registry' stamp: 'bgf 5/15/2006 14:18'!
newNameFor: anObject
	^TObjectID for: anObject! !

!TIsland methodsFor: 'registry' stamp: 'das 9/6/2005 19:11'!
register: anObject
	^self register: anObject name: (self newNameFor: anObject)! !

!TIsland methodsFor: 'registry' stamp: 'ar 2/1/2006 17:08'!
register: anObject name: aName
	| ref |
	true ifTrue:["@@@ Temporary sanity check @@@"
		ref := self lookup: aName ifAbsent:[nil].
		(ref == nil or:[(ref valueOn: self) == anObject]) 
			ifFalse:[self error: 'Duplicate ID'].
	].
	ref := anObject 
		isFarRef ifTrue:[anObject] 
		ifFalse:[self asFarRef: anObject].
	ref myName ifNil:[ref myName: aName].
	mutex critical:[nameMap at: aName put: ref].
	^ref! !

!TIsland methodsFor: 'accessing' stamp: 'ar 11/8/2005 18:44'!
roots
	"Answer the roots for this island"
	| roots |
	self flag: #fixMe.
	"I don't quite understand why I have added the following GC but it does not work without it. I need to review it and either add a lengthy comment why we need it or (even better) add a comment explaining what needs to be fixed so we can get by withou it. There really doesn't seem to be much of a reason for it here."
	Smalltalk garbageCollect.
	roots := Array streamContents:[:s|
		s nextPut: self. "always myself first"
		exports do:[:each| s nextPut: each].
		nameMap do:[:each| s nextPut: each].
	].
	exports growTo: (Set sizeFor: roots size).
	^roots! !

!TIsland methodsFor: 'comparing' stamp: 'jrd 8/19/2009 22:58'!
sameIslandAs: aTFarRef

	^aTFarRef hasIsland: self! !

!TIsland methodsFor: 'event loop' stamp: 'ar 9/17/2005 19:57'!
scheduleMessage: msgSend at: timeStamp
	"Schedule the given message deltaSeconds into the future"
	msgSend time: timeStamp.
	"FIXME: Wrapping below means that if we have an overflow messages scheduled after the overflow will be executed *before* messages scheduled earlier. It won't lead to any collisions (this would require wrap-around within a time slot) but it still is a problem since it may cause unpredictable effects on the code."
	msgSend sequenceNumber: (sequence := sequence + 1 bitAnd: 16r3FFFFFFF).
	messageQueue add: msgSend.! !

!TIsland methodsFor: 'event loop' stamp: 'ar 11/3/2005 12:08'!
scheduleMessage: msgSend future: deltaMSecs
	"Schedule the given message deltaSeconds into the future"
	^self scheduleMessage: msgSend at: time + deltaMSecs! !

!TIsland methodsFor: 'actions' stamp: 'jrd 9/16/2009 21:29'!
sendMessage: message
	"@@@@@FIXME: We have a potential problem here...sometimes this can be
	reached when the controller is nil, because a snapshot is in progress. There
	is a mutex that is in play during the snapshot, but it is in the controller....so
	we can't reach it. And even if we could, the datapath for sending the message
	is completely independent of the sync...it would just be a meaningless delay.
	Until we can come up with a better plan, we busy wait with a delay until we
	have a controller, then forward the message. We wait 100 ms as that
	seems to be close to a lower limit on the sync time."
	"This is one of those places where I'm not sure how it ever worked for more than
	5 min in a row...."
	"Possible better plans: have farrefs refer to the controller rather than the
	island, or maybe both. Have an islandID->controller lookup table someplace
	public in case the controller ivar is nil "

	| ctrl |
	ctrl := controller.
	[ ctrl isNil ]
		whileTrue: [
			(Delay forMilliseconds: 100) wait.
			ctrl := controller
		].
	^ctrl sendMessage: message
! !

!TIsland methodsFor: 'accessing' stamp: 'ar 9/16/2005 00:34'!
time
	^time! !

!TIsland methodsFor: 'accessing' stamp: 'ar 9/16/2005 00:34'!
time: msecs
	time := msecs.! !

!TIsland methodsFor: 'accessing' stamp: 'ar 11/3/2005 12:07'!
totalSeconds
	"Answer the total seconds expired"
	^time asInteger // 1000! !

!TIsland methodsFor: 'registry' stamp: 'jrd 2/22/2012 15:37'!
unregister: anObject
	"exports is a weak dictionary, but nameMap is forever.
	Until we have garbage collection, we need a way to manually remove stuff.
	One example is #makePointer, which is just used in the dynamic extent of an event handler.
	"
	nameMap removeKey: (self asFarRef: anObject) myName ifAbsent: ["@@@@supicious"].
! !

!TIsland methodsFor: 'event loop' stamp: 'ar 9/7/2006 15:40'!
unscheduleMessage: aMessage
	"Unschedule the given message, e.g., remove it from the message queue."
	^(self unscheduleMessageMatching:[:msg| msg == aMessage]) notNil! !

!TIsland methodsFor: 'event loop' stamp: 'ar 9/7/2006 15:39'!
unscheduleMessageMatching: aBlock
	"Unschedule any single message matching aBlock."
	1 to: messageQueue size do:[:i|
		(aBlock value: (messageQueue at: i)) ifTrue:[^messageQueue removeAt: i].
	].
	^nil! !

!Object methodsFor: '*Croquet-Kernel' stamp: 'sn 2/23/2018 21:50'!
croquetFork: aSelector at: aPriority
	| aProcess |
	aProcess :=  [self perform: aSelector] croquetForkAt: aPriority.
	aProcess name: self class name,'>>', aSelector.
	^aProcess! !

!Object methodsFor: '*Croquet-Kernel' stamp: 'das 1/6/2006 16:08'!
fork: aSelector
	^self fork: aSelector at: Processor activePriority! !

!Object methodsFor: '*Croquet-Kernel' stamp: 'das 1/6/2006 16:07'!
fork: aSelector at: aPriority
	| aProcess |
	aProcess :=  [self perform: aSelector] forkAt: aPriority.
	aProcess name: self class name,'>>', aSelector.
	^aProcess! !

!Object methodsFor: '*Croquet-Kernel' stamp: 'sn 2/26/2018 16:31'!
future
	^(TFutureMaker new) 
		setDeltaMSecs: 0.0 target: self! !

!Object methodsFor: '*Croquet-Kernel' stamp: 'ar 11/3/2005 11:57'!
future: deltaMSecs
	^(TFutureMaker new)
		setDeltaMSecs: deltaMSecs target: self! !

!Object methodsFor: '*Croquet-Kernel' stamp: 'ar 11/3/2005 11:57'!
futureDo: aSelector at: deltaMSecs args: args
	"Send a message deltaSeconds into the future"
	| msgSend |
	msgSend := (TMessageSend new)
		receiver: self;
		selector: aSelector;
		arguments: args.
	Processor activeIsland scheduleMessage: msgSend future: deltaMSecs.
	^nil! !

!Object methodsFor: '*Croquet-Kernel' stamp: 'bgf 5/14/2006 14:01'!
futureSend: aSelector at: deltaMSecs args: args
	"As an interim measure we will not allow internal future messages to create promises. It is *not* certain if that's the right thing to do or not but it avoids some nasty representation issues, such as the magical conversion from promise to value. For external futures this is okay because the promise is represented as far ref and resolving it simply fills in the value but for internal futures we would need an extra promise representation (TPromise comes to mind) which needs to be resolved explicitly (e.g., via become: or so). Since this is ugly we will try to avoid it."
	^self futureDo: aSelector at: deltaMSecs args: args

"The old code:
	| msgSend promise id |
	promise := TPromise new.
	promise id: (id := (TObjectID for: 'Promise').
	promise island: Processor activeIsland.
	promise register.
	promise runScript: #onResolved: when:{Processor activeIsland. id asSymbol}.
	msgSend := (TMessageSend new)
		receiver: self;
		selector: aSelector;
		arguments: args.
	msgSend id: id.
	Processor activeIsland scheduleMessage: msgSend future: deltaMSecs.
	^promise
"! !

!Object methodsFor: '*Croquet-Kernel' stamp: 'ar 9/17/2005 19:40'!
send: aBlock
	^aBlock value: self! !

Object subclass: #TARC4Cypher
	instanceVariableNames: 'm x y'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TARC4Cypher commentStamp: '<historical>' prior: 0!
I represent the RC4 stream cypher.!

!TARC4Cypher class methodsFor: 'instance creation' stamp: 'ar 11/5/2005 22:54'!
key: aKey
	^self new key: aKey! !

!TARC4Cypher class methodsFor: 'instance creation' stamp: 'jrd 11/22/2008 20:38'!
key: aKey drop: numBytes
	^(self new)
		key: aKey ;
		drop: numBytes ;
		yourself! !

!TARC4Cypher class methodsFor: 'tests' stamp: 'ar 11/5/2005 22:54'!
test
	"ARC4 tests vectors from OpenSSL (crypto/rc4/rc4test.c)"
	"TRC4Cypher test"
	| keys inputs outputs buffer cypher |
	keys := #(
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16rEF 16r01 16r23 16r45)
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16rEF 16r01 16r23 16r45)
	).
	inputs := #(
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 
		 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00
		 16r00 16r00 16r00 16r00)
		(16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
	).
	outputs := #(
		(16r75 16rB7 16r87 16r80 16r99 16rE0 16rC5 16r96)
		(16r74 16r94 16rC2 16rE7 16r10 16r4B 16r08 16r79)
		(16rDE 16r18 16r89 16r41 16rA3 16r37 16r5D 16r3A)
		(16rD6 16rA1 16r41 16rA7 16rEC 16r3C 16r38 16rDF
		 16rBD 16r61 16r5A 16r11 16r62 16rE1 16rC7 16rBA
		 16r36 16rB6 16r78 16r58)
		(16r66 16rA0 16r94 16r9F 16r8A 16rF7 16rD6 16r89
		 16r1F 16r7F 16r83 16r2B 16rA8 16r33 16rC0 16r0C
		 16r89 16r2E 16rBE 16r30 16r14 16r3C 16rE2 16r87
		 16r40 16r01 16r1E 16rCF)
		(16rD6 16rA1 16r41 16rA7 16rEC 16r3C 16r38 16rDF 16rBD 16r61)
	).
	1 to: keys size do:[:i|
		buffer := (inputs at: i) copy.
		cypher := self key: (keys at: i).
		cypher process: buffer.
		buffer = (outputs at: i) ifFalse:[self error: 'Encryption failure'].
		cypher := self key: (keys at: i).
		cypher process: buffer.
		buffer = (inputs at: i) ifFalse:[self error: 'Decryption failure'].
	].! !

!TARC4Cypher methodsFor: 'cypher' stamp: 'jrd 11/22/2008 20:39'!
drop: numBytes

	self next: numBytes! !

!TARC4Cypher methodsFor: 'cypher' stamp: 'ar 3/28/2006 11:15'!
initKey: key
	| j k a keyLen |
	keyLen := key size.
	x := y := 0.
	m := ByteArray new: 256.
	0 to: 255 do:[:i| m at: i+1 put: i].
	j := k := 0.
	0 to: 255 do:[:i|
		a := m at: i+1.
		j := (j + a + (key at: k + 1)) bitAnd: 255.
		m at: i+1 put: (m at: j+1). m at: j+1 put: a.
		(k := k + 1) >= keyLen ifTrue:[k := 0].
	].
! !

!TARC4Cypher methodsFor: 'cypher' stamp: 'ar 3/28/2006 11:15'!
key: key
	| buffer |
	self initKey: key.
	"Always strip off the first 1024 bytes of the key stream since they are 
	non-uniformely distributed and can been used to attack the cypher."
	buffer := ByteArray new: 32.
	1 to: 32 do:[:i| self process: buffer from: 1 to: buffer size].! !

!TARC4Cypher methodsFor: 'streaming' stamp: 'ar 1/28/2006 21:45'!
next
	"Answer the next byte from this random stream"
	^(self nextInto: (ByteArray new: 1)) at: 1! !

!TARC4Cypher methodsFor: 'streaming' stamp: 'ar 1/28/2006 21:47'!
next: nBytes
	"Store the next n bytes from the random key stream in aCollection"
	^self next: nBytes into: (ByteArray new: nBytes) startingAt: 1! !

!TARC4Cypher methodsFor: 'streaming' stamp: 'ar 1/28/2006 21:47'!
next: nBytes into: aCollection
	"Store the next n bytes from the random key stream in aCollection"
	^self next: nBytes into: aCollection startingAt: 1! !

!TARC4Cypher methodsFor: 'streaming' stamp: 'ar 1/28/2006 21:47'!
next: n into: aCollection startingAt: start
	"Store the next n bytes from the random key stream in aCollection"
	^self process: aCollection from: start to: start + n - 1! !

!TARC4Cypher methodsFor: 'streaming' stamp: 'ar 1/28/2006 21:47'!
nextInto: aCollection
	"Store the next n bytes from the random key stream in aCollection"
	^self next: aCollection size into: aCollection startingAt: 1! !

!TARC4Cypher methodsFor: 'streaming' stamp: 'ar 1/28/2006 21:47'!
nextInto: aCollection startingAt: start
	"Store the next n bytes from the random key stream in aCollection"
	^self next: aCollection size - start + 1 into: aCollection startingAt: start! !

!TARC4Cypher methodsFor: 'cypher' stamp: 'ar 3/29/2006 01:09'!
primitiveARC4Process: buffer from: startIndex to: stopIndex m: mm x: xx y: yy
	"Run the primitive. If successful, the return value is x@y"
	<primitive: 'primitiveARC4Transform' module: 'CroquetPlugin'>
	^nil "indicates failure"
! !

!TARC4Cypher methodsFor: 'cypher' stamp: 'ar 11/5/2005 23:15'!
process: buffer
	^self process: buffer from: 1 to: buffer size! !

!TARC4Cypher methodsFor: 'cypher' stamp: 'ar 3/26/2006 19:58'!
process: buffer from: startIndex to: stopIndex
	"Process the given buffer"
	| a b mask xAtY |
	"Try the primitive first"
	xAtY := self primitiveARC4Process: buffer from: startIndex to: stopIndex m: m x: x y: y.
	xAtY ifNotNil:[
		x := xAtY x.
		y := xAtY y.
		^buffer
	].
	startIndex to: stopIndex do:[:i|
		x := (x + 1) bitAnd: 255.
		a := m at: x+1.
		y := (y + a) bitAnd: 255.
		b := m at: y+1.
		m at: x+1 put: b.
		m at: y+1 put: a.
		mask := m at: ((a + b) bitAnd: 255) + 1.
		buffer at: i put: ((buffer at: i) bitXor: mask).
	].
	^buffer! !

!TARC4Cypher methodsFor: 'streaming' stamp: 'ar 1/28/2006 22:17'!
seedVector: aByteArray
	"Seed the random key vector with aByteArray"
	^self key: aByteArray! !

TARC4Cypher subclass: #TARC4DropNCypher
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TARC4DropNCypher commentStamp: '<historical>' prior: 0!
I represent the RC4 stream cypher with the first n bytes dropped to avoid a weak key attack.
SCAN database name: RC4-drop(n). !

!TARC4DropNCypher class methodsFor: 'as yet unclassified' stamp: 'jrd 11/23/2008 01:52'!
conservativeKey: aKey
	"3072 is named the conservative drop size in the SCAN entry for RC4-drop(n)"

	^self key: aKey drop: 3072! !

!TARC4DropNCypher class methodsFor: 'as yet unclassified' stamp: 'jrd 11/23/2008 01:52'!
key: aKey
	"768 is the default drop size in the SCAN entry for RC4-drop(n)"

	^self key: aKey drop: 768! !

Object subclass: #TCryptoRandom
	instanceVariableNames: 'pool entropy rotate poolIndex lastTick lastDelta'
	classVariableNames: 'Default PoolBits PoolMask PoolWords TAP1 TAP2 TAP3 TAP4 TAP5 TwistTable'
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TCryptoRandom commentStamp: '<historical>' prior: 0!
I am a cryptographically strong random number generator. My implementation is based on code found at http://www.cs.berkeley.edu/~daw/rnd/linux-rand.!

!TCryptoRandom class methodsFor: 'arc4' stamp: 'ar 1/28/2006 21:23'!
arc4: input key: key
	| output |
	output := input shallowCopy.
	(TARC4Cypher key: key) process: output.
	^output! !

!TCryptoRandom class methodsFor: 'arc4' stamp: 'ar 1/28/2006 21:22'!
arc4Test
	"ARC4 tests vectors from OpenSSL (crypto/rc4/rc4test.c)"
	"self arc4Test"
	| keys inputs outputs result |
	keys := #(
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16rEF 16r01 16r23 16r45)
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16rEF 16r01 16r23 16r45)
	).
	inputs := #(
		(16r01 16r23 16r45 16r67 16r89 16rAB 16rCD 16rEF)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 
		 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00
		 16r00 16r00 16r00 16r00)
		(16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78 16r9A 16rBC 16rDE 16rF0
		 16r12 16r34 16r56 16r78)
		(16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00)
	).
	outputs := #(
		(16r75 16rB7 16r87 16r80 16r99 16rE0 16rC5 16r96)
		(16r74 16r94 16rC2 16rE7 16r10 16r4B 16r08 16r79)
		(16rDE 16r18 16r89 16r41 16rA3 16r37 16r5D 16r3A)
		(16rD6 16rA1 16r41 16rA7 16rEC 16r3C 16r38 16rDF
		 16rBD 16r61 16r5A 16r11 16r62 16rE1 16rC7 16rBA
		 16r36 16rB6 16r78 16r58)
		(16r66 16rA0 16r94 16r9F 16r8A 16rF7 16rD6 16r89
		 16r1F 16r7F 16r83 16r2B 16rA8 16r33 16rC0 16r0C
		 16r89 16r2E 16rBE 16r30 16r14 16r3C 16rE2 16r87
		 16r40 16r01 16r1E 16rCF)
		(16rD6 16rA1 16r41 16rA7 16rEC 16r3C 16r38 16rDF 16rBD 16r61)
	).
	1 to: keys size do:[:i|
		result := self arc4: (inputs at: i) key: (keys at: i).
		result = (outputs at: i) ifFalse:[self error: 'Encryption failure'].
		result := self arc4: result key: (keys at: i).
		result = (inputs at: i) ifFalse:[self error: 'Decryption failure'].
	].! !

!TCryptoRandom class methodsFor: 'accessing' stamp: 'ar 9/15/2005 02:59'!
default
	^Default ifNil:[Default := self new].! !

!TCryptoRandom class methodsFor: 'accessing' stamp: 'ar 10/13/2005 18:53'!
default: aRandom
	Default := aRandom.! !

!TCryptoRandom class methodsFor: 'entropy' stamp: 'ar 3/28/2006 17:32'!
gatherEntropy: aByteArray
	"Fill buffer's worth of data with entropy. Return true if a high-quality source of entropy has been used, false if we had to fall back to a low-quality source."
	<primitive: 'primitiveGatherEntropy' module: 'CroquetPlugin'>
	self default nextInto: aByteArray.
	^false! !

!TCryptoRandom class methodsFor: 'secure hash' stamp: 'ar 1/5/2006 13:45'!
hashMessage: aStringOrByteArray
	"This is an entry vector which we can use to dispatch to various hash algorithms"
	^self md5HashMessage: aStringOrByteArray! !

!TCryptoRandom class methodsFor: 'secure hash' stamp: 'ar 1/5/2006 13:45'!
hashStream: aStream
	"This is an entry vector which we can use to dispatch to various hash algorithms"
	^self md5HashStream: aStream! !

!TCryptoRandom class methodsFor: 'class initialization' stamp: 'ar 9/14/2005 21:57'!
initialize
	"TCryptoRandom initialize"
	PoolWords := 128.    "Power of 2 - note that this is 32-bit words"
	PoolMask := PoolWords-1.
	PoolBits := PoolWords * 32.

	"The pool is stirred with a primitive polynomial of degree 128
	over GF(2), namely x^128 + x^99 + x^59 + x^31 + x^9 + x^7 + 1."
	TAP1 := 99.     "The polynomial taps"
	TAP2 := 59.
	TAP3 := 31.
	TAP4 := 9.
	TAP5 := 7.

	TwistTable := #(
		16r00000000	16r3B6E20C8	16r76DC4190	16r4DB26158
		16rEDB88320	16rD6D6A3E8	16r9B64C2B0	16rA00AE278
	).

	Default := self new.
! !

!TCryptoRandom class methodsFor: 'md5' stamp: 'ar 11/5/2005 20:05'!
md5HashMessage: aStringOrByteArray
	"TCryptoRandom md5HashMessage: 'foo'"
	^ self md5HashStream: (ReadStream on: aStringOrByteArray asByteArray)
! !

!TCryptoRandom class methodsFor: 'md5' stamp: 'ar 3/26/2006 19:48'!
md5HashStream: aStream
	"TCryptoRandom md5HashStream: (ReadStream on: 'foo')"
	| start buffer bytes sz n words hash |
	hash := WordArray with: 16r67452301 with: 16rEFCDAB89 with: 16r98BADCFE with: 16r10325476.
	words := WordArray new: 16.
	buffer := ByteArray new: 64.
	start := aStream position.
	[aStream atEnd] whileFalse: [
		bytes := aStream nextInto: buffer.
		(bytes size < 64 or:[aStream atEnd]) ifTrue:[
			sz := bytes size.
			buffer replaceFrom: 1 to: sz with: bytes startingAt: 1.
			buffer from: sz+1 to: buffer size put: 0.
			sz < 56 ifTrue:[
				buffer at: sz + 1 put: 128. "trailing bit"
			] ifFalse:[
				"not enough room for the length, so just pad this one, then..."
				sz < 64 ifTrue:[buffer at: sz + 1 put: 128].
				1 to: 16 do:[:i| words at: i put: (buffer unsignedLongAt: i*4-3 bigEndian: false)].
				self md5Transform: words hash: hash.
				"process one additional block of padding ending with the length"
				buffer atAllPut: 0.
				sz = 64 ifTrue: [buffer at: 1 put: 128].
			].
			"Fill in the final 8 bytes with the 64-bit length in bits."
			n := (aStream position - start) * 8.
			7 to: 0 by: -1 do:[:i| buffer at: (buffer size - i) put: ((n bitShift: 7-i*-8) bitAnd: 255)].
		].
		1 to: 16 do:[:i| words at: i put: (buffer unsignedLongAt: i*4-3 bigEndian: false)].
		self md5Transform: words hash: hash.
	].
	bytes := TObjectID nilObjectID.
	bytes unsignedLongAt: 1 put: (hash at: 4) bigEndian: true.
	bytes unsignedLongAt: 5 put: (hash at: 3) bigEndian: true.
	bytes unsignedLongAt: 9 put: (hash at: 2) bigEndian: true.
	bytes unsignedLongAt: 13 put: (hash at: 1) bigEndian: true.
	^bytes! !

!TCryptoRandom class methodsFor: 'md5' stamp: 'ar 11/5/2005 20:05'!
md5Test
	"TCryptoRandom md5Test"
	(self md5HashMessage: 'a') asInteger = 16r0CC175B9C0F1B6A831C399E269772661
		ifFalse: [ self error: 'failed'].
	(self md5HashMessage: 'abc') asInteger = 16r900150983CD24FB0D6963F7D28E17F72
		ifFalse: [ self error: 'failed'].
	(self md5HashMessage: 'message digest') asInteger = 16rF96B697D7CB7938D525A2F31AAF161D0
		ifFalse: [ self error: 'failed'].
	(self md5HashMessage:
		'abcdefghijklmnopqrstuvwxyz') asInteger = 16rC3FCD3D76192E4007DFB496CCA67E13B
		ifFalse: [ self error: 'failed'].
	(self md5HashMessage:
	'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789') asInteger =
		16rD174AB98D277D9F5A5611C2C9F419D9F
		ifFalse: [ self error: 'failed'].
	(self md5HashMessage:
	'12345678901234567890123456789012345678901234567890123456789012345678901234567890') asInteger = 		16r57EDF4A22BE3C955AC49DA2E2107B67A
		ifFalse: [ self error: 'failed'].
! !

!TCryptoRandom class methodsFor: 'md5' stamp: 'ar 3/26/2006 19:53'!
md5Transform: in hash: hash
	"This adds the incoming words to the existing hash"
	| a b c d |
	<primitive: 'primitiveMD5Transform' module: 'CroquetPlugin'>
	a := hash at: 1.
	b := hash at: 2.
	c := hash at: 3.
	d := hash at: 4.

	a := self step1: a x: b y: c z: d data: (in at:  1) add: 16rD76AA478 shift: 7.
	d := self step1: d x: a y: b z: c data: (in at:  2) add: 16rE8C7B756 shift: 12.
	c := self step1: c x: d y: a z: b data: (in at:  3) add: 16r242070DB shift: 17.
	b := self step1: b x: c y: d z: a data: (in at:  4) add: 16rC1BDCEEE shift: 22.
	a := self step1: a x: b y: c z: d data: (in at:  5) add: 16rF57C0FAF shift:  7.
	d := self step1: d x: a y: b z: c data: (in at:  6) add: 16r4787C62A shift: 12.
	c := self step1: c x: d y: a z: b data: (in at:  7) add: 16rA8304613 shift: 17.
	b := self step1: b x: c y: d z: a data: (in at:  8) add: 16rFD469501 shift: 22.
	a := self step1: a x: b y: c z: d data: (in at:  9) add: 16r698098D8 shift:  7.
	d := self step1: d x: a y: b z: c data: (in at: 10) add: 16r8B44F7AF shift: 12.
	c := self step1: c x: d y: a z: b data: (in at: 11) add: 16rFFFF5BB1 shift: 17.
	b := self step1: b x: c y: d z: a data: (in at: 12) add: 16r895CD7BE shift: 22.
	a := self step1: a x: b y: c z: d data: (in at: 13) add: 16r6B901122 shift:  7.
	d := self step1: d x: a y: b z: c data: (in at: 14) add: 16rFD987193 shift: 12.
	c := self step1: c x: d y: a z: b data: (in at: 15) add: 16rA679438E shift: 17.
	b := self step1: b x: c y: d z: a data: (in at: 16) add: 16r49B40821 shift: 22.

	a := self step2: a x: b y: c z: d data: (in at:  2) add: 16rF61E2562 shift:  5.
	d := self step2: d x: a y: b z: c data: (in at:  7) add: 16rC040B340 shift:  9.
	c := self step2: c x: d y: a z: b data: (in at: 12) add: 16r265E5A51 shift: 14.
	b := self step2: b x: c y: d z: a data: (in at:  1) add: 16rE9B6C7AA shift: 20.
	a := self step2: a x: b y: c z: d data: (in at:  6) add: 16rD62F105D shift:  5.
	d := self step2: d x: a y: b z: c data: (in at: 11) add: 16r02441453 shift:  9.
	c := self step2: c x: d y: a z: b data: (in at: 16) add: 16rD8A1E681 shift: 14.
	b := self step2: b x: c y: d z: a data: (in at:  5) add: 16rE7D3FBC8 shift: 20.
	a := self step2: a x: b y: c z: d data: (in at: 10) add: 16r21E1CDE6 shift:  5.
	d := self step2: d x: a y: b z: c data: (in at: 15) add: 16rC33707D6 shift:  9.
	c := self step2: c x: d y: a z: b data: (in at:  4) add: 16rF4D50D87 shift: 14.
	b := self step2: b x: c y: d z: a data: (in at:  9) add: 16r455A14ED shift: 20.
	a := self step2: a x: b y: c z: d data: (in at: 14) add: 16rA9E3E905 shift:  5.
	d := self step2: d x: a y: b z: c data: (in at:  3) add: 16rFCEFA3F8 shift:  9.
	c := self step2: c x: d y: a z: b data: (in at:  8) add: 16r676F02D9 shift: 14.
	b := self step2: b x: c y: d z: a data: (in at: 13) add: 16r8D2A4C8A shift: 20.

	a := self step3: a x: b y: c z: d data: (in at:  6) add: 16rFFFA3942 shift:  4.
	d := self step3: d x: a y: b z: c data: (in at:  9) add: 16r8771F681 shift: 11.
	c := self step3: c x: d y: a z: b data: (in at: 12) add: 16r6D9D6122 shift: 16.
	b := self step3: b x: c y: d z: a data: (in at: 15) add: 16rFDE5380C shift: 23.
	a := self step3: a x: b y: c z: d data: (in at:  2) add: 16rA4BEEA44 shift:  4.
	d := self step3: d x: a y: b z: c data: (in at:  5) add: 16r4BDECFA9 shift: 11.
	c := self step3: c x: d y: a z: b data: (in at:  8) add: 16rF6BB4B60 shift: 16.
	b := self step3: b x: c y: d z: a data: (in at: 11) add: 16rBEBFBC70 shift: 23.
	a := self step3: a x: b y: c z: d data: (in at: 14) add: 16r289B7EC6 shift:  4.
	d := self step3: d x: a y: b z: c data: (in at:  1) add: 16rEAA127FA shift: 11.
	c := self step3: c x: d y: a z: b data: (in at:  4) add: 16rD4EF3085 shift: 16.
	b := self step3: b x: c y: d z: a data: (in at:  7) add: 16r04881D05 shift: 23.
	a := self step3: a x: b y: c z: d data: (in at: 10) add: 16rD9D4D039 shift:  4.
	d := self step3: d x: a y: b z: c data: (in at: 13) add: 16rE6DB99E5 shift: 11.
	c := self step3: c x: d y: a z: b data: (in at: 16) add: 16r1FA27CF8 shift: 16.
	b := self step3: b x: c y: d z: a data: (in at:  3) add: 16rC4AC5665 shift: 23.

	a := self step4: a x: b y: c z: d data: (in at:  1) add: 16rF4292244 shift:  6.
	d := self step4: d x: a y: b z: c data: (in at:  8) add: 16r432AFF97 shift: 10.
	c := self step4: c x: d y: a z: b data: (in at: 15) add: 16rAB9423A7 shift: 15.
	b := self step4: b x: c y: d z: a data: (in at:  6) add: 16rFC93A039 shift: 21.
	a := self step4: a x: b y: c z: d data: (in at: 13) add: 16r655B59C3 shift:  6.
	d := self step4: d x: a y: b z: c data: (in at:  4) add: 16r8F0CCC92 shift: 10.
	c := self step4: c x: d y: a z: b data: (in at: 11) add: 16rFFEFF47D shift: 15.
	b := self step4: b x: c y: d z: a data: (in at:  2) add: 16r85845DD1 shift: 21.
	a := self step4: a x: b y: c z: d data: (in at:  9) add: 16r6FA87E4F shift:  6.
	d := self step4: d x: a y: b z: c data: (in at: 16) add: 16rFE2CE6E0 shift: 10.
	c := self step4: c x: d y: a z: b data: (in at:  7) add: 16rA3014314 shift: 15.
	b := self step4: b x: c y: d z: a data: (in at: 14) add: 16r4E0811A1 shift: 21.
	a := self step4: a x: b y: c z: d data: (in at:  5) add: 16rF7537E82 shift:  6.
	d := self step4: d x: a y: b z: c data: (in at: 12) add: 16rBD3AF235 shift: 10.
	c := self step4: c x: d y: a z: b data: (in at:  3) add: 16r2AD7D2BB shift: 15.
	b := self step4: b x: c y: d z: a data: (in at: 10) add: 16rEB86D391 shift: 21.

	a := (a + (hash at: 1)) bitAnd: 16rFFFFFFFF. hash at: 1 put: a.
	b := (b + (hash at: 2)) bitAnd: 16rFFFFFFFF. hash at: 2 put: b.
	c := (c + (hash at: 3)) bitAnd: 16rFFFFFFFF. hash at: 3 put: c.
	d := (d + (hash at: 4)) bitAnd: 16rFFFFFFFF. hash at: 4 put: d.

	^hash! !

!TCryptoRandom class methodsFor: 'md5' stamp: 'ar 11/5/2005 20:07'!
rotate: value by: amount
	"Rotate value left by amount"
	| lowMask highMask |
	lowMask := (1 bitShift: 32-amount) - 1.
	highMask := 16rFFFFFFFF - lowMask.
	^((value bitAnd: lowMask) bitShift: amount) + 
		((value bitAnd: highMask) bitShift: amount-32)! !

!TCryptoRandom class methodsFor: 'md5' stamp: 'ar 11/5/2005 20:06'!
step1: w x: x y: y z: z data: data add: add shift: s
	"First step in MD5 transformation"
	| f result |
	f := z bitXor: (x bitAnd: (y bitXor: z)).
	result := w + f + data + add.
	result := self rotate: result by: s.
	^result + x bitAnd: 16rFFFFFFFF! !

!TCryptoRandom class methodsFor: 'md5' stamp: 'ar 11/5/2005 20:06'!
step2: w x: x y: y z: z data: data add: add shift: s
	"First step in MD5 transformation"
	| f result |
	f := y bitXor: (z bitAnd: (x bitXor: y)).
	result := w + f + data + add.
	result := self rotate: result by: s.
	^result + x bitAnd: 16rFFFFFFFF! !

!TCryptoRandom class methodsFor: 'md5' stamp: 'ar 11/5/2005 20:06'!
step3: w x: x y: y z: z data: data add: add shift: s
	"First step in MD5 transformation"
	| f result |
	f := (x bitXor: y) bitXor: z.
	result := w + f + data + add.
	result := self rotate: result by: s.
	^result + x bitAnd: 16rFFFFFFFF! !

!TCryptoRandom class methodsFor: 'md5' stamp: 'ar 11/5/2005 20:06'!
step4: w x: x y: y z: z data: data add: add shift: s
	"First step in MD5 transformation"
	| f result |
	f := y bitXor: (x bitOr: (z  bitXor: 16rFFFFFFFF)).
	result := w + f + data + add.
	result := self rotate: result by: s.
	^result + x bitAnd: 16rFFFFFFFF! !

!TCryptoRandom methodsFor: 'accessing' stamp: 'ar 9/15/2005 02:55'!
addEntropyWord: value
	"This method adds a word to the entropy pool."
	| w mask |
	w := self rotate: value by: rotate.
	poolIndex := poolIndex - 1.
	poolIndex > 0 ifTrue:[
		rotate := (rotate + 7) bitAnd: 31.
	] ifFalse:[
		"At the beginning of the pool, add an extra 7 bits
		rotation, so that successive passes spread the
		input bits across the pool evenly."
		rotate := (rotate + 14) bitAnd: 31.
		poolIndex := PoolWords.
	].
	"XOR in the various taps"
	mask := PoolMask.
	w := w bitXor: (pool at: (poolIndex + TAP1 bitAnd: mask) + 1).
	w := w bitXor: (pool at: (poolIndex + TAP2 bitAnd: mask) + 1).
	w := w bitXor: (pool at: (poolIndex + TAP3 bitAnd: mask) + 1).
	w := w bitXor: (pool at: (poolIndex + TAP4 bitAnd: mask) + 1).
	w := w bitXor: (pool at: (poolIndex + TAP5 bitAnd: mask) + 1).
	w := w bitXor: (pool at: poolIndex).
	pool at: poolIndex put: ((w bitShift: -3) bitXor: (TwistTable at: (w bitAnd: 7) + 1)).
! !

!TCryptoRandom methodsFor: 'accessing' stamp: 'ar 9/16/2005 01:34'!
addTimerRandomness: value
	"Add entropy to the pool by using timing delays"
	^self addTimerRandomness: value ticks: Processor activeIsland millisecondClockValue.! !

!TCryptoRandom methodsFor: 'accessing' stamp: 'ar 9/16/2005 01:34'!
addTimerRandomness: value ticks: nextTick
	"Add entropy to the pool by using timing delays"
	| delta delta2 bits |

	self addEntropyWord: nextTick.
	self addEntropyWord: value.

	"Calculate the number of bits randomness we probably added"

	delta := nextTick - lastTick.	lastTick := nextTick.
	delta2 := delta - lastDelta.	lastDelta := delta.

	delta < 0 ifTrue:[delta := 0 - delta].
	delta2 < 0 ifTrue:[delta2 := 0 - delta2].
	delta > delta2 ifTrue:[delta := delta2].

	bits := (delta bitShift: -1) highBit min: 12.
	entropy := entropy + bits.
	entropy > PoolBits ifTrue:[entropy := PoolBits].! !

!TCryptoRandom methodsFor: 'initialize' stamp: 'ar 9/15/2005 03:53'!
initialize
	super initialize.
	pool := WordArray new: PoolWords.
	poolIndex := 1.
	rotate := 0.
	entropy := 0.
	lastTick := 0.
	lastDelta := 0.
	self addTimerRandomness: Processor activeIsland totalSeconds.
! !

!TCryptoRandom methodsFor: 'accessing' stamp: 'ar 1/28/2006 21:46'!
next
	"Answer the next byte from this random stream"
	^(self nextInto: (ByteArray new: 1)) at: 1! !

!TCryptoRandom methodsFor: 'accessing' stamp: 'ar 9/15/2005 02:15'!
next: nBytes
	^self next: nBytes into: (ByteArray new: nBytes) startingAt: 1! !

!TCryptoRandom methodsFor: 'accessing' stamp: 'ar 9/15/2005 02:15'!
next: nBytes into: aCollection
	^self next: nBytes into: aCollection startingAt: 1! !

!TCryptoRandom methodsFor: 'accessing' stamp: 'ar 3/26/2006 19:50'!
next: nBytes into: aCollection startingAt: startIndex
	"This is the heart of the matter. Fetch nBytes of random input."
	| hash words buf n sz index |
	buf := ByteArray new: 16.
	n := nBytes.
	index := startIndex.
	[n > 0] whileTrue:[
		"Start the md5 hash tail"
		hash := WordArray with: 16r67452301 with: 16rEFCDAB89 with: 16r98BADCFE with: 16r10325476.
		words := WordArray new: 16.
		1 to: PoolWords by: 16 do:[:i|
			words replaceFrom: 1 to: 16 with: pool startingAt: i.
			TCryptoRandom md5Transform: words hash: hash.
		].
		"Modify pool so next hash will produce different results"
		self addEntropyWord: (hash at: 1).
		self addEntropyWord: (hash at: 2).
		self addEntropyWord: (hash at: 3).
		self addEntropyWord: (hash at: 4).
		"Run the MD5 Transform one more time, since we want
		to add at least minimal obscuring of the inputs to addEntropyWord:"
		TCryptoRandom md5Transform: pool hash: hash.
		buf unsignedLongAt: 1 put: (hash at: 1) bigEndian: false.
		buf unsignedLongAt: 5 put: (hash at: 2) bigEndian: false.
		buf unsignedLongAt: 9 put: (hash at: 3) bigEndian: false.
		buf unsignedLongAt: 13 put: (hash at: 4) bigEndian: false.
		sz := n min: 16. n := n - sz.
		aCollection replaceFrom: index to: index + sz-1 with: buf startingAt: 1.
		index := index + sz.
	].
	^aCollection! !

!TCryptoRandom methodsFor: 'accessing' stamp: 'ar 9/15/2005 02:52'!
nextInto: aCollection
	^self next: aCollection size into: aCollection startingAt: 1! !

!TCryptoRandom methodsFor: 'accessing' stamp: 'ar 9/15/2005 02:52'!
nextInto: aCollection startingAt: start
	^self next: aCollection size - start + 1 into: aCollection startingAt: start! !

!TCryptoRandom methodsFor: 'private' stamp: 'ar 11/5/2005 20:08'!
rotate: value by: amount
	"Rotate value left by amount"
	| lowMask highMask |
	lowMask := (1 bitShift: 32-amount) - 1.
	highMask := 16rFFFFFFFF - lowMask.
	^((value bitAnd: lowMask) bitShift: amount) + 
		((value bitAnd: highMask) bitShift: amount-32)! !

Object subclass: #TFarRef
	instanceVariableNames: 'myIsland myValue myName myController destroyed'
	classVariableNames: 'AllFarRefs'
	poolDictionaries: ''
	category: 'Croquet-Kernel-Objects'!

!TFarRef commentStamp: '<historical>' prior: 0!
A Croquet specific FarRef. Does ->NOT<- allow immediate messages to pass through automatically. Use only in conjunction with #future or #send, e.g.,
	ref := island send avatar.
	ref future translation: 100@0@100.
	trans := ref send:[:obj| obj translation]. "<- this will likely NOT be 100@0@100"
!

!TFarRef class methodsFor: 'class initialization' stamp: 'ar 9/17/2005 20:27'!
allFarRefs
	^AllFarRefs! !

!TFarRef class methodsFor: 'accessing' stamp: 'jrd 12/11/2009 00:03'!
destroyRefsForController: controller! !

!TFarRef class methodsFor: 'accessing' stamp: 'das 9/6/2005 17:08'!
howToPassAsArgument
	"passByProxy: by definition"
	^#passByProxy:! !

!TFarRef class methodsFor: 'class initialization' stamp: 'ar 9/17/2005 20:34'!
initialize
	"TFarRef initialize"
	AllFarRefs := TFarRefRegistry new.
	AllFarRefs addAll: self allSubInstances.! !

!TFarRef methodsFor: 'comparing' stamp: 'das 9/6/2005 16:44'!
= anObject
	"Answer a SmallInteger whose value is related to the receiver's identity.
	May be overridden, and should be overridden in any classes that define = "
	^self == anObject! !

!TFarRef methodsFor: 'accessing' stamp: 'das 2/1/2006 19:45'!
asFarRef

	^ self.! !

!TFarRef methodsFor: 'copying' stamp: 'ar 1/13/2006 14:43'!
clone
	"Do not allow generically copying far refs"
	^self! !

!TFarRef methodsFor: 'future sends' stamp: 'das 12/15/2005 12:23'!
color

	^ self send:[ :sp | sp color.].
! !

!TFarRef methodsFor: 'accessing' stamp: 'mtf 7/16/2010 18:48'!
controller

	"pvtController may answer a FarRef to the currently active island, which is incorrect; sync sending yourself will resolve the invalid FarRef if so, and do nothing if not."
	^self pvtController yourself! !

!TFarRef methodsFor: 'copying' stamp: 'ar 1/13/2006 14:43'!
copy
	"Do not allow generically copying far refs"
	^self! !

!TFarRef methodsFor: 'private' stamp: 'jrd 1/5/2010 20:35'!
debuggingIsland

	^myIsland! !

!TFarRef methodsFor: 'initialize' stamp: 'das 12/3/2005 10:46'!
destroy

	myIsland := nil.
	myValue := nil.
	myName := nil.! !

!TFarRef methodsFor: 'private' stamp: 'das 9/6/2005 17:04'!
exportOnIslandWriter: writer
	^writer exportFarRef: self! !

!TFarRef methodsFor: 'message sends' stamp: 'jrd 12/2/2009 19:54'!
futureDo: aSelector at: deltaMSecs args: args
	"Send a message deltaSeconds into the future"
	| msgData encoder |
	deltaMSecs = 0 ifFalse:[^self error:'Cannot send timed external messages'].
	myName ifNil:[^self error: 'This object has no name'].
	msgData := TMessageData new.
	msgData receiver: myName.
	msgData selector: aSelector.
	encoder := TMessageEncoder new.
	encoder island: self island.
	encoder nextPut: args.
	msgData arguments: encoder contents.
	msgData time: deltaMSecs.
	self pvtController sendMessage: msgData.
	^nil! !

!TFarRef methodsFor: 'message sends' stamp: 'sn 2/23/2018 22:18'!
futureSend: aSelector at: deltaMSecs args: args
	"Send a message deltaSeconds into the future"
	| msgData encoder id promised | 

	deltaMSecs = 0 ifFalse:[^self error:'Cannot send timed external messages'].
	myName ifNil:[^self error: 'This object has no name'].
	id := (TObjectID for: aSelector).
	promised := TFarRef new.
	promised privateValue: promised island: myIsland.
	"@@@@: Check this - is this a potential race condition??? @@@@"
	myIsland register: promised name: id shallowCopy.
	msgData := TMessageData new.
	msgData receiver: myName.
	msgData selector: aSelector.
	encoder := TMessageEncoder new.
	encoder island: self island.
	encoder nextPut: args.
	msgData arguments: encoder contents.
	msgData id: id.
	msgData time: deltaMSecs.
	self pvtController sendMessage: msgData.
	^promised! !

!TFarRef methodsFor: 'message sends' stamp: 'bgf 6/26/2006 23:03'!
get: aSymbol

	"Shorthand version of foo send:[:x| x aSymbol].
	Returns nil on any error (usually indicates absence of the field.)
	Note: this differs from Object/CObject get: in that
	the selector is invoked, rather than treated as an instance variable name."

	[[
		^ self syncSend: aSymbol withArguments: #()
	 ] on: MessageNotUnderstood do: [ :ex |

		"Silently return nil - there's no such selector."
		ex return

	]] on: Error do: [ :ex |

		"Something's wrong... report the problem, nondisruptively,
          since we don't want to pollute this Island with other's problems."

		Transcript cr; show: 'Error in TFarRef#get: ', ex description.
		ex return
	].
	^nil! !

!TFarRef methodsFor: 'future sends' stamp: 'das 12/15/2005 12:19'!
globalMatrixOut

	^ self send:[ :vp | vp globalMatrixOut.].
! !

!TFarRef methodsFor: 'comparing' stamp: 'jrd 8/19/2009 22:58'!
hasIsland: otherMyIsland

	^myIsland = otherMyIsland! !

!TFarRef methodsFor: 'comparing' stamp: 'das 9/6/2005 16:44'!
hash
	"Answer a SmallInteger whose value is related to the receiver's identity.
	May be overridden, and should be overridden in any classes that define = "
	^self identityHash! !

!TFarRef methodsFor: 'initialize' stamp: 'ar 9/17/2005 20:34'!
initialize
	AllFarRefs add: self.! !

!TFarRef methodsFor: 'accessing' stamp: 'das 9/6/2005 17:05'!
isFarRef
	^true! !

!TFarRef methodsFor: 'testing' stamp: 'ar 11/8/2005 17:59'!
isResolved
	"Answer whether this far ref is resolved"
	^myValue ~~ self "e.g., as long as I am unresolved I represent myself"! !

!TFarRef methodsFor: 'future sends' stamp: 'das 12/15/2005 12:18'!
isSpace

	^ self send:[ :vp | vp isSpace.].
! !

!TFarRef methodsFor: 'accessing' stamp: 'das 9/6/2005 17:05'!
island
	"Answer the receiver's island"
	^(myIsland asFarRef: myIsland) valueOn: Processor activeIsland! !

!TFarRef methodsFor: 'private' stamp: 'das 9/6/2005 15:08'!
myName
	^myName! !

!TFarRef methodsFor: 'private' stamp: 'das 9/6/2005 15:08'!
myName: aString
	myName := aString! !

!TFarRef methodsFor: 'private' stamp: 'jrd 8/19/2009 20:51'!
objectForDataStream: refStream
    "Return an object to store on an external data stream."
	myName ifNil:[^self error: 'This object has no name'].
"@@@@@//@@@@@ Should we unpack them into farrefs if the proxy is for another island?"
	(refStream sameIslandAs: self)
		ifFalse: [self log warn: ('{Potential problem with TFarRef serialization: island mismatch: ' , myName printString , '}')].
    ^(TObjectProxy new: 16)
		replaceFrom: 1 to: 16 with: myName startingAt: 1;
		yourself! !

!TFarRef methodsFor: 'scripting' stamp: 'ar 9/8/2005 11:56'!
on: eventName notify: anObserver
	"Relay events from my target to any observers"
	^self syncSend: #on:notify: withArguments: {eventName. anObserver}! !

!TFarRef methodsFor: 'scripting' stamp: 'ar 9/8/2005 11:57'!
on: eventName unNotify: anObserver
	"Relay events from my target to any observers"
	^self syncSend: #on:unNotify: withArguments: {eventName. anObserver}! !

!TFarRef methodsFor: 'private' stamp: 'ar 2/2/2006 15:25'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut:$(.
		self isResolved
			ifTrue:[aStream nextPutAll: (self get: #printString)]
			ifFalse:[aStream nextPutAll: 'unresolved'].
	aStream nextPut:$).! !

!TFarRef methodsFor: 'private' stamp: 'jrd 12/3/2009 15:39'!
privateValue: anObject island: anIsland
	(myIsland == nil or:[myIsland == anIsland])
		ifFalse:[^self error:'Cannot change farRef island'].
	(myValue == nil or:[myValue == self or:[myValue == anObject]]) 
		ifFalse:[^self error:'Cannot change farRef value'].
	myValue := anObject.
	myIsland := anIsland.
	self pvtController.! !

!TFarRef methodsFor: 'private' stamp: 'jrd 12/3/2009 15:39'!
pvtController

	^myController
		ifNil: [
			(myController := myIsland controller)
		]! !

!TFarRef methodsFor: 'future sends' stamp: 'das 1/19/2006 14:20'!
rearrange: ogl

	^ self send:[ :vp | vp rearrange: ogl.].
! !

!TFarRef methodsFor: 'future sends' stamp: 'das 12/13/2005 18:21'!
renderAsOverlay: ogl

	^ self send:[ :vp | vp renderAsOverlay: ogl.].
! !

!TFarRef methodsFor: 'future sends' stamp: 'das 12/15/2005 12:23'!
renderSpace: ogl port: portal depth: depth ghostFrame: ghost

	^ self send:[ :sp | sp renderSpace: ogl port: portal depth: depth ghostFrame: ghost.].
! !

!TFarRef methodsFor: 'future sends' stamp: 'das 3/22/2006 13:05'!
renderSubspace: ogl

	^ self send:[ :space | space renderSubspace: ogl.].
! !

!TFarRef methodsFor: 'future sends' stamp: 'das 12/13/2005 18:22'!
renderView: ogl overlay: ov

	^ self send:[ :vp | vp renderView: ogl overlay: ov.].
! !

!TFarRef methodsFor: 'future sends' stamp: 'das 12/15/2005 12:16'!
root

	^ self send:[ :vp | vp root.].
! !

!TFarRef methodsFor: 'comparing' stamp: 'jrd 8/19/2009 22:57'!
sameIslandAs: aTFarRef

	^aTFarRef hasIsland: myIsland ! !

!TFarRef methodsFor: 'message sends' stamp: 'ar 9/17/2005 19:38'!
send: aBlock
	"Evaluate aBlock with a FarRef to the receiver which allows relaying immediate messages"
	^aBlock value: (TMessageMaker new setTarget: self)! !

!TFarRef methodsFor: 'copying' stamp: 'ar 1/13/2006 14:43'!
shallowCopy
	"Do not allow generically copying far refs"
	^self! !

!TFarRef methodsFor: 'private' stamp: 'das 9/7/2005 11:11'!
syncReturn: result from: dstIsland to: srcIsland
	^(IslandArgumentCopier new from: dstIsland to: srcIsland) pass: result.! !

!TFarRef methodsFor: 'message sends' stamp: 'ar 9/16/2005 00:19'!
syncSend: aMessage
	^self syncSend: aMessage selector withArguments: aMessage arguments! !

!TFarRef methodsFor: 'private' stamp: 'sn 2/23/2018 23:34'!
syncSend: selector withArguments: arguments 
	"Send the selector, aSymbol, to the receiver with arguments in arguments."
	| copier args dstIsland srcIsland result pp arg |
	myValue == self ifTrue:[^self error:'Cannot send sync messages to unresolved far refs'].
	srcIsland := Processor activeIsland.
	dstIsland := myIsland.
	srcIsland == dstIsland 
		ifTrue:[^myValue perform: selector withArguments: arguments].

	copier := nil.
	args := arguments shallowCopy.
	1 to: arguments size do:[:i|
		arg := arguments at: i.
		(IslandArgumentCopier isPassByIdentity: arg) ifFalse:[
			copier ifNil:[copier := IslandArgumentCopier new from:  srcIsland to: dstIsland].
			arg := copier pass: arg.
		].
		args at: i put: arg.
	].
	
	result := Processor activeProcess activateIslandWithEscape: myIsland
		during: [myValue perform: selector withArguments: args].
	^self syncReturn: result from: dstIsland to: srcIsland
	
	"pp := Processor activeProcess.

	pp island: myIsland.
	[[
		result := myValue perform: selector withArguments: args]
		ensure:[pp island: srcIsland]]
		on: Exception do: [:ex |
			pp island: srcIsland. ex outer.
			pp island: dstIsland. ex resume].
	^self syncReturn: result from: dstIsland to: srcIsland"! !

!TFarRef methodsFor: 'accessing' stamp: 'ar 11/8/2005 18:07'!
value
	^self syncReturn: myValue from: myIsland to: Processor activeIsland! !

!TFarRef methodsFor: 'accessing' stamp: 'das 9/6/2005 17:05'!
valueOn: anIsland
	"Answer the object/value I would have on the given island"
	anIsland == myIsland ifTrue:[^myValue] ifFalse:[^self].! !

!TFarRef methodsFor: 'accessing' stamp: 'das 9/6/2005 17:05'!
varValueOn: anIsland
	"Answer the object/value I would have on the given island.
	This version is specific for variables handling copy on read."
	anIsland == myIsland ifTrue:[^myValue].
	"Do a quick check for #passByProxy: ... "
	myValue class howToPassAsArgument == #passByProxy: ifTrue:[^self].
	" ... but otherwise use slow, generic mechanism."
	anIsland isFarRef ifTrue:[^self error: 'Cannot compute value on far island'].
	^self syncReturn: myValue from: myIsland to: anIsland! !

!TFarRef methodsFor: 'waiting' stamp: 'das 1/6/2006 15:32'!
wait
	"Wait unconditionally for this reference to resolve."
	| sema |
	sema := Semaphore new.
	self whenResolved:[sema signal].
	sema wait.! !

!TFarRef methodsFor: 'waiting' stamp: 'das 1/6/2006 15:33'!
waitTimeoutMSecs: msecs
	"Wait for at most the given number of milliseconds for this promise to resolve. Answer true if it is resolved, false otherwise."
	| sema |
	sema := Semaphore new.
	self whenResolved:[sema signal].
	Delay timeoutSemaphore: sema afterMSecs: msecs.
	^self isResolved! !

!TFarRef methodsFor: 'initialize' stamp: 'das 1/6/2006 14:14'!
whenResolved: aBlock
	"Evaluate aBlock when I am resolved"
	| script |
	aBlock numArgs <= 1 ifFalse:[self error: 'Must be zero or one argument block'].
	self isResolved ifTrue:[
		aBlock numArgs = 0 ifTrue:[^aBlock value].
		aBlock numArgs = 1 ifTrue:[^aBlock value: self].
	].
	script := self island runScript:[
		aBlock numArgs = 0 ifTrue:[aBlock value].
		aBlock numArgs = 1 ifTrue:[aBlock value: self].
		self island stopScript: script.
	] when: {self island. myName asSymbol}.! !

Object subclass: #TFarRefRegistry
	instanceVariableNames: 'array tally flag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TFarRefRegistry commentStamp: '<historical>' prior: 0!
A registry for managing TFarRefs!

!TFarRefRegistry methodsFor: 'accessing' stamp: 'ar 9/17/2005 20:33'!
add: anObject
	"Include anObject as one of the receiver's elements, but only if
	not already present. Answer farRef."
	| index |
	anObject ifNil: [self error: 'Sets cannot meaningfully contain nil as an element'].
	index := self findElementOrFlag: anObject.
	(array at: index) == flag ifTrue:[self atNewIndex: index put: anObject].
	^anObject! !

!TFarRefRegistry methodsFor: 'accessing' stamp: 'ar 9/17/2005 20:34'!
addAll: aCollection
	aCollection do:[:each| self add: each].! !

!TFarRefRegistry methodsFor: 'accessing' stamp: 'jrd 12/22/2009 19:16'!
allForIsland: island do: block

	self
		do: [ :each |
			(each sameIslandAs: island )
				ifTrue: [ block value: each ]
		].! !

!TFarRefRegistry methodsFor: 'converting' stamp: 'ar 9/17/2005 20:32'!
asArray
	^Array streamContents:[:s| self do:[:ref| s nextPut: ref]].! !

!TFarRefRegistry methodsFor: 'private' stamp: 'ar 9/17/2005 20:31'!
atNewIndex: index put: anObject
	array at: index put: anObject.
	tally := tally + 1.
	self fullCheck.! !

!TFarRefRegistry methodsFor: 'accessing' stamp: 'ar 9/17/2005 20:37'!
do: aBlock
	| obj |
	tally = 0 ifTrue:[^self].
	1 to: array size do:[:i|
		((obj := array at: i) == nil or:[obj == flag]) ifFalse:[aBlock value: obj].
	].! !

!TFarRefRegistry methodsFor: 'private' stamp: 'ar 9/17/2005 20:30'!
findElementOrFlag: anObject
	"Answer the index of a first slot containing either a nil (indicating an empty slot) or an element that matches the given object. Answer the index of that slot or zero. Fail if neither a match nor an empty slot is found."

	| index |
	[(index := self scanFor: anObject) = 0] whileTrue:[self grow].
	^index! !

!TFarRefRegistry methodsFor: 'private' stamp: 'ar 9/17/2005 20:32'!
fullCheck
	"Keep array at least 1/4 free for decent hash behavior"
	| minFree nowFree |
	minFree := array size // 4 max: 1.
	nowFree := array size - tally.
	nowFree > minFree ifTrue:[^self].
	"Make some room by dumping niled out elements"
	self growTo: array size.
	nowFree := array size - tally.
	nowFree > minFree ifTrue:[
		"Try to shrink if we have very items"
		tally < minFree ifTrue:[self shrink].
		^self].
	"Really grow"
	self grow.! !

!TFarRefRegistry methodsFor: 'accessing' stamp: 'ar 9/17/2005 20:31'!
grow
	"Grow the elements array and reinsert the old elements"
	self growTo: array size + self growSize! !

!TFarRefRegistry methodsFor: 'private' stamp: 'ar 9/17/2005 20:30'!
growSize
	^ array size max: 2! !

!TFarRefRegistry methodsFor: 'private' stamp: 'jrd 12/29/2009 21:35'!
growTo: anInteger
	"Grow the elements array and reinsert the old elements"

	| oldElements |
	oldElements := array.
	array := WeakArray new: anInteger.
	array atAllPut: flag.
	tally := 0.
	oldElements do:[:obj |
		(obj == flag or: [obj == nil]) ifFalse: [
			array at: (self findElementOrFlag: obj ) put: obj.
			tally := tally + 1.
		]
	].! !

!TFarRefRegistry methodsFor: 'initialize' stamp: 'ar 9/17/2005 20:28'!
initialize
	flag := Object new.
	tally := 0.
	array := WeakArray new: (Set sizeFor: 40).
	array atAllPut: flag.! !

!TFarRefRegistry methodsFor: 'accessing' stamp: 'ar 9/17/2005 20:31'!
rehash
	self growTo: array size.! !

!TFarRefRegistry methodsFor: 'accessing' stamp: 'jrd 12/29/2009 22:54'!
removeAllForIsland: island
	"I'm not sure if I should use flag or nil to force remove the elements."

	| obj |
	tally = 0
		ifTrue:[^self].
	1
		to: array size
		do:[ :i |
			((obj := array at: i) == nil or:[obj == flag])
				ifFalse:[
					(obj sameIslandAs: island)
						ifTrue: [
							array at: i put: flag .
							tally := tally - 1.
						 ]
				].
		].! !

!TFarRefRegistry methodsFor: 'private' stamp: 'jrd 12/29/2009 22:48'!
scanFor: anObject
	"Scan the key array for the first slot containing either a _flag_ (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."
"Ignore nil elements...they are from the weak array finalization and will get cleaned up during the grow process"

	| finish hash start element |
	finish := array size.
	finish > 4096
		ifTrue: [hash := anObject identityHash * (finish // 4096)]
		ifFalse: [hash := anObject identityHash].
	start := (hash \\ array size) + 1.

	"Search from (hash mod size) to the end."
	start to: finish do:[:index | 
		(element := array at: index) ifNotNil:[
			(element == flag or: [element == anObject])
				ifTrue: [^ index ]]].

	"Search from 1 to where we started."
	1 to: start-1 do:[:index | 
		(element := array at: index) ifNotNil:[
			(element == flag or: [element == anObject])
				ifTrue: [^ index ]]].

	^ 0  "No match AND no empty slot"! !

!TFarRefRegistry methodsFor: 'accessing' stamp: 'ar 9/17/2005 20:31'!
shrink
	"Shrink the elements array and reinsert the old elements"
	self growTo: (Set sizeFor: tally)! !

Object subclass: #TLogger
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

TLogger class
	instanceVariableNames: 'mutex'!

!TLogger commentStamp: '<historical>' prior: 0!
TLogger logs important messages to a file named 'Croquet.log'.!

TLogger class
	instanceVariableNames: 'mutex'!

!TLogger class methodsFor: 'class initialization' stamp: 'mtf 1/22/2010 16:54'!
initialize
	"TLogger initialize"
	mutex := TMutex new.! !

!TLogger class methodsFor: 'logging' stamp: 'jrd 12/15/2008 15:56'!
log: aString
	"Log aString"
	| mtime |
	self withLogFileDo:[:file|
		mtime := DateAndTime now.
		file cr.
		mtime printYMDOn: file.
		file nextPut: $T.
		mtime printHMSOn: file.
		file
			nextPutAll: ' : 	'.
		file cr; nextPutAll: aString.
		file flush.
	].! !

!TLogger class methodsFor: 'logging' stamp: 'mtf 1/22/2010 16:55'!
withLogFileDo: aBlock
	| file filePath |
	filePath := FileDirectory pathFrom: {FileDirectory default pathName . 'log'. 'Croquet.log' }.
	mutex critical:[
		file := [FileStream fileNamed: filePath] 
			on: Error
			do:[:ex| ex return: nil].
		file
			ifNotNil:[
				[
					[
						file setToEnd.
						aBlock value: file
					]
						on: Error
						do:[:ex| ex return]
				] ensure:[file close]
			].
	].
! !

Object subclass: #TMessageData
	instanceVariableNames: 'receiver selector arguments sender msgId time sid'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Messages'!

!TMessageData commentStamp: '<historical>' prior: 0!
I represent a single message (datagram).

Instance variables:
	receiver	<TObjectID>	The receiver's id.
	selector		<Symbol>	The message selector.
	arguments	<ByteArray>	The message arguments (encoded!!)
	sender		<TObjectID>	The sender's (machine) id.
	msgId		<TObjectID>	The id for this message (to be signaled upon completion)
	time		<Float>		The time (in seconds) when this message should be run
	sid			<Integer>	The sequence id for this message!

!TMessageData class methodsFor: 'islands' stamp: 'das 9/7/2005 16:05'!
howToPassAsArgument
	^#passByClone:! !

!TMessageData methodsFor: 'comparing' stamp: 'mtf 7/21/2010 13:21'!
<= other
	^ self sid <= other sid! !

!TMessageData methodsFor: 'comparing' stamp: 'mtf 7/21/2010 13:11'!
= other
	^ (self sid = other sid
		and: [self time = other time]
		and: [self id = other id]
		and: [self sender = other sender])
		and: [self receiver = other receiver]
		and: [self selector = other selector]
		and: [self arguments = other arguments]
! !

!TMessageData methodsFor: 'accessing' stamp: 'das 9/6/2005 11:12'!
arguments
	^arguments! !

!TMessageData methodsFor: 'accessing' stamp: 'das 9/6/2005 16:51'!
arguments: args
	arguments := args! !

!TMessageData methodsFor: 'read/write' stamp: 'das 9/6/2005 18:12'!
asByteArray
	"Convert me into a datagram"
	| strm |
	strm := WriteStream on: (ByteArray new: arguments size + 64).
	self writeOn: strm.
	^strm contents! !

!TMessageData methodsFor: 'comparing' stamp: 'mtf 7/21/2010 13:18'!
hash
	^ ((((((self sid hash
		bitXor: self time hash)
		bitXor: self id hash)
		bitXor: self sender hash)
		bitXor: self receiver hash)
		bitXor: self selector hash)
		bitXor: self arguments hash)
! !

!TMessageData methodsFor: 'accessing' stamp: 'ar 11/8/2005 17:04'!
id
	^msgId! !

!TMessageData methodsFor: 'accessing' stamp: 'ar 11/8/2005 17:04'!
id: tid
	msgId := tid! !

!TMessageData methodsFor: 'initialize' stamp: 'ar 10/18/2005 17:02'!
initialize
	sid := 0.
	time := 0.0.
	msgId := TObjectID nilObjectID.! !

!TMessageData methodsFor: 'read/write' stamp: 'ar 1/5/2006 13:04'!
readFrom: aStream
	| count |
	receiver := aStream nextInto: (TObjectID nilObjectID).
	count := aStream nextNumber: 4.
	selector := (aStream next: count) asString asSymbol.
	count := aStream nextNumber: 4.
	arguments := aStream next: count.
	sender := aStream nextInto: (TObjectID nilObjectID).
	msgId := aStream nextInto: (TObjectID nilObjectID).
	time := Float basicNew: 2.
	time basicAt: 1 put: (aStream nextNumber: 4).
	time basicAt: 2 put: (aStream nextNumber: 4).
	sid := aStream nextNumber: 4.! !

!TMessageData methodsFor: 'accessing' stamp: 'das 9/6/2005 11:12'!
receiver
	^receiver! !

!TMessageData methodsFor: 'accessing' stamp: 'das 9/6/2005 16:51'!
receiver: rcvr
	receiver := rcvr.! !

!TMessageData methodsFor: 'accessing' stamp: 'das 9/6/2005 11:12'!
selector
	^selector! !

!TMessageData methodsFor: 'accessing' stamp: 'das 9/6/2005 16:51'!
selector: aSymbol
	selector := aSymbol.! !

!TMessageData methodsFor: 'accessing' stamp: 'das 11/4/2005 11:09'!
sender
	^sender! !

!TMessageData methodsFor: 'accessing' stamp: 'das 11/4/2005 11:10'!
sender: aTObjectID
	sender := aTObjectID! !

!TMessageData methodsFor: 'accessing' stamp: 'das 9/6/2005 11:12'!
sid
	^sid! !

!TMessageData methodsFor: 'accessing' stamp: 'das 9/6/2005 11:12'!
sid: aNumber
	sid := aNumber.! !

!TMessageData methodsFor: 'accessing' stamp: 'das 9/6/2005 11:12'!
time
	^time! !

!TMessageData methodsFor: 'accessing' stamp: 'ar 10/18/2005 17:02'!
time: msecs
	time := msecs asFloat.! !

!TMessageData methodsFor: 'read/write' stamp: 'das 11/4/2005 11:15'!
writeOn: aStream
	aStream nextPutAll: receiver.
	aStream nextNumber: 4 put: selector size.
	aStream nextPutAll: selector asByteArray.
	aStream nextNumber: 4 put: arguments size.
	aStream nextPutAll: arguments.
	aStream nextPutAll: sender.
	aStream nextPutAll: msgId.
	aStream nextNumber: 4 put: (time basicAt: 1).
	aStream nextNumber: 4 put: (time basicAt: 2).
	aStream nextNumber: 4 put: sid.
! !

Object subclass: #TMutex
	instanceVariableNames: 'semaphore owner'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Objects'!

!TMutex commentStamp: 'mtf 7/23/2010 10:50' prior: 0!
A Mutex is a light-weight MUTual EXclusion object being used when two or more processes need to access a shared resource concurrently. A Mutex grants ownership to a single process and will suspend any other process trying to aquire the mutex while in use. Waiting processes are granted access to the mutex in the order the access was requested.

A single process can acquire a mutex multiple times, unlike a Semaphore. This returns hello:
|m| m := TMutex new. m critical: [m critical: ['hello']]

but this just deadlocks:
|m| m := Semaphore forMutualExclusion. m critical: [m critical: ['hello']]

Instance variables:
	semaphore	<Semaphore>		The (primitive) semaphore used for synchronization.
	owner		<Process>		The process owning the mutex.!

!TMutex methodsFor: 'mutual exclusion' stamp: 'ar 11/3/2005 05:02'!
critical: aBlock
	"Evaluate aBlock protected by the receiver."
	| activeProcess |
	activeProcess := Processor activeProcess.
	activeProcess == owner ifTrue:[^aBlock value].
	^semaphore critical:[
		owner := activeProcess.
		aBlock ensure:[owner := nil]].! !

!TMutex methodsFor: 'initialize' stamp: 'ar 11/3/2005 04:35'!
initialize
	semaphore := Semaphore forMutualExclusion.! !

Object subclass: #TMutexSet
	instanceVariableNames: 'array'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Objects'!

!TMutexSet commentStamp: '<historical>' prior: 0!
A TMutexSet aquires all of the mutexes inside it.!

!TMutexSet class methodsFor: 'instance creation' stamp: 'ar 11/3/2005 05:07'!
withAll: mutexList
	^self new withAll: mutexList! !

!TMutexSet methodsFor: 'mutual exclusion' stamp: 'ar 11/3/2005 05:09'!
critical: aBlock
	"Evaluate aBlock aquiring all mutexes"
	^self pvtCritical: aBlock startingAt: 1! !

!TMutexSet methodsFor: 'private' stamp: 'ar 11/3/2005 05:09'!
pvtCritical: aBlock startingAt: index
	| mutex |
	index > array size ifTrue:[^aBlock value].
	mutex := array at: index.
	^mutex critical:[self pvtCritical: aBlock startingAt: index+1].! !

!TMutexSet methodsFor: 'initialize' stamp: 'ar 11/3/2005 05:07'!
withAll: mutexList
	array := mutexList.! !

Object subclass: #TObject
	instanceVariableNames: 'myProperties myCobaltObjectProperties'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Objects'!

!TObject commentStamp: '<historical>' prior: 0!
A base class for dealing with some unpleasantness of registries etc.

Instance variables:
	myProperties	<TObjectProperties>	The extra properties for this object.
!

TObject subclass: #TIslandController
	instanceVariableNames: 'status log island eventQueue eventLoop networkQueue facets mutex senderID gcData messageRecorder'
	classVariableNames: 'AccessProtect Registry'
	poolDictionaries: ''
	category: 'Croquet-Kernel-Objects'!

!TIslandController commentStamp: 'mtf 7/21/2010 15:22' prior: 0!
I control an island's execution.

Instance variables:
	status			<Symbol>			A symbol indicating the status of the controller:
		#closed			- Unconnected
		#connecting		- Trying to connect to a router
		#authenticating	- Authenticating a user
		#ready				- Connected and ready to use.
		#invalid			- Something went really, really wrong.
	log				<Stream>			A log stream for debug messages.
	island			<TIsland>			The island I am responsible for.
	eventQueue	<SharedQueue of MessageSend>	The shared queue holding the messages for
															the event loop.
	eventLoop		<Process>		The process handling messages in the queue.
	networkQueue	<SharedQueue of TMessageData>	The queue of messages to send into the island
	facets			<Dictionary>		A dictionary storing two kinds of data on facets, depending on key:
		<TObjectID -> MessageSend>	Maps a facet ID to a message and reciever to invoke
		<Symbol -> TObjectID>			Maps a convenient name to a facet ID (not necessarily
											one I understand)
	mutex			<TMutex>			The mutex used to gain exclusive access to an island.
	senderID		<TObjectID>		This controller's sender ID.!

!TIslandController class methodsFor: 'accessing' stamp: 'ar 1/30/2006 23:38'!
controllerWithID: id
	^AccessProtect critical:[Registry at: id ifAbsent:[nil]]! !

!TIslandController class methodsFor: 'accessing' stamp: 'ar 1/31/2006 00:01'!
initialize
	"TRemoteController initialize"
	AccessProtect := TMutex new.
	Registry := WeakValueDictionary new.! !

!TIslandController methodsFor: 'facets' stamp: 'jrd 11/22/2008 17:18'!
addFacetCalled: symbolName sending: selector

	| msgFacet |
	msgFacet := self addMessageFacet: selector.
	facets at: symbolName put: msgFacet.
	^msgFacet
! !

!TIslandController methodsFor: 'facets' stamp: 'bgf 5/14/2006 14:08'!
addMessageFacet: aFacet
	"Add the given message send as a facet under the provided 128 bit id"
	| fid msgSend |
	aFacet isSymbol 
		ifTrue:[msgSend := self messageFacetNamed: aFacet]
		ifFalse:[msgSend := aFacet].
	fid := TObjectID for: 'Facet'.
	facets at: fid put: msgSend.
	^fid! !

!TIslandController methodsFor: 'actions' stamp: 'sn 2/23/2018 21:21'!
advanceTo: newTime
  
	island ifNil:[^nil]. "we are probably still sync-ing"
	self processMessages. "process all the messages thus far"
	island send:[:obj| obj advanceTo: newTime].! !

!TIslandController methodsFor: 'actions' stamp: 'ar 11/3/2005 05:53'!
beServer
	"Act as a server, e.g., a participant trusted to replicate an island."
	^self subclassResponsibility! !

!TIslandController methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:34'!
cacheManager
	"Backstop"
	^nil! !

!TIslandController methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:34'!
cacheManager: aCacheMgr
	"Backstop"
	^nil! !

!TIslandController methodsFor: 'locking' stamp: 'ar 11/3/2005 05:03'!
critical: aBlock
	"Aquire exclusive access to the island for the duration of aBlock"
	^mutex critical: aBlock! !

!TIslandController methodsFor: 'initialize' stamp: 'jrd 1/7/2010 19:53'!
destroy
	AccessProtect critical:[Registry removeKey: senderID ifAbsent:[]].
	status := #invalid.
	self stopEventLoop.
	self withGcDataDo: [ : aGcData | aGcData controller: nil].
	gcData := nil.
! !

!TIslandController methodsFor: 'accessing' stamp: 'ar 9/15/2005 20:52'!
eventQueue
	^eventQueue! !

!TIslandController methodsFor: 'accessing' stamp: 'ar 9/14/2005 13:40'!
facets
	^facets! !

!TIslandController methodsFor: 'accessing' stamp: 'ar 9/14/2005 13:40'!
facets: aMap
	facets := aMap! !

!TIslandController methodsFor: 'accessing' stamp: 'jrd 1/7/2010 18:45'!
gcData

	^gcData! !

!TIslandController methodsFor: 'accessing' stamp: 'jrd 1/7/2010 18:44'!
gcData: aGcData

	gcData := aGcData

	! !

!TIslandController methodsFor: 'actions' stamp: 'das 11/3/2005 19:41'!
heartbeat: tickPeriod
	"HEARTBEAT: Request period update information from the router"
	^self subclassResponsibility! !

!TIslandController methodsFor: 'initialize' stamp: 'jrd 1/7/2010 19:50'!
initialize
	super initialize.
	status := #invalid. "default - subclasses should change status if appropriate"
	senderID := TObjectID for: self.
	AccessProtect critical:[Registry at: senderID put: self].
	mutex := TMutex new.
	eventQueue := SharedQueue new.
	networkQueue := SharedQueue new.
	facets := Dictionary new.
	self startEventLoop.
! !

!TIslandController methodsFor: 'accessing' stamp: 'das 9/7/2005 16:10'!
island
	^island! !

!TIslandController methodsFor: 'accessing' stamp: 'jrd 3/12/2010 21:18'!
island: anIsland
	self shutdownIslandSignalHandlers.
	island := anIsland.
	self setupIslandSignalHandlers.! !

!TIslandController methodsFor: 'initialize' stamp: 'jrd 5/4/2009 01:20'!
islandClass

	^TIsland! !

!TIslandController methodsFor: 'actions' stamp: 'ar 11/3/2005 05:54'!
join
	"JOIN: Request to receive messages for this island."
	^self subclassResponsibility! !

!TIslandController methodsFor: 'actions' stamp: 'ar 11/3/2005 05:54'!
leave
	"LEAVE: Request no more messages for this island."
	^self subclassResponsibility! !

!TIslandController methodsFor: 'printing' stamp: 'jrd 4/30/2009 13:03'!
logControllerTMessage: msg! !

!TIslandController methodsFor: 'actions' stamp: 'das 11/4/2005 11:19'!
measureRoundtrip: msg! !

!TIslandController methodsFor: 'facets' stamp: 'ar 9/16/2005 10:18'!
messageFacetNamed: aSelector
	"Create and answer an invokable facet for the given selector"
	^MessageSend receiver: self selector: aSelector! !

!TIslandController methodsFor: 'locking' stamp: 'ar 11/3/2005 05:05'!
mutex
	^mutex! !

!TIslandController methodsFor: 'initialize' stamp: 'jrd 3/12/2010 21:19'!
newIsland
	"Create and install new island for this controller."
	| newIsland |
	newIsland := self islandClass new.
	newIsland send:[:obj| obj controller: self].
	self island: newIsland.
	^island! !

!TIslandController methodsFor: 'printing' stamp: 'ar 1/4/2006 20:36'!
printExtrasOn: aStream
	aStream nextPut:$(.
	aStream nextPutAll: status asString.
	aStream nextPut:$).
! !

!TIslandController methodsFor: 'printing' stamp: 'ar 1/4/2006 14:58'!
printOn: aStream
	super printOn: aStream.
	self printExtrasOn: aStream.
! !

!TIslandController methodsFor: 'actions' stamp: 'ar 3/17/2006 18:42'!
processMessages
	"Process pending messages for this island"
	| nextMsg |
	island ifNil:[^nil]. "we are probably still sync-ing"
	self critical:[
		"Get the next message from the (concurrent) network queue"
		[nextMsg := networkQueue nextOrNil.
		nextMsg == nil] whileFalse:[
			"And have the island decode, schedule, and update to that message"
			island send:[:obj| obj decodeScheduleAndExecute: nextMsg].
		].
	].
! !

!TIslandController methodsFor: 'initialize' stamp: 'jrd 12/21/2009 13:41'!
reflectReplicateValues: valueRef  into:  ocRef

	ocRef future addLast: ( self senderID copy  ->  valueRef value).! !

!TIslandController methodsFor: 'debugging' stamp: 'jrd 12/22/2009 19:29'!
relatedRefOwnersToIsland

	| myFarRefOwners |
	myFarRefOwners := IdentitySet new.
	TFarRef allFarRefs
		allForIsland: island
		do: [ :each |
			myFarRefOwners addAll: (each inboundPointers)
		].
	^myFarRefOwners! !

!TIslandController methodsFor: 'snapshot' stamp: 'jrd 3/12/2010 14:42'!
restore: data
	"Restore a snapshot of an island"
	| reader newIsland |
	reader := TSnapshotReader new.
	newIsland := reader readObjectFrom: data readStream.
	newIsland send:[:obj| obj controller: self].
	^newIsland


"
	| newIsland refId |
	newIsland := super restore: data.
	restoringFromCache = true \ifTrue: [
		refId := newIsland myName.
		newIsland send: [:i | i compactNameMap: refId.].
		self setRouterTime: (newIsland get: #time).
	].
	^newIsland."! !

!TIslandController methodsFor: 'snapshot' stamp: 'jrd 1/24/2012 16:21'!
restore: data newIslandId: islandID
	"Restore a snapshot of an island"
	| reader newIsland oldId idsToChange |
	reader := TSnapshotReader new.
	newIsland := reader readObjectFrom: data readStream.
	newIsland send:[:obj| obj controller: self].
	self setupIslandSignalHandlers.
	islandID
		ifNotNil: [
			oldId := newIsland get: #id.
			idsToChange := TObjectID allInstances select: [ :eachTObjectID | eachTObjectID == oldId].
			idsToChange do: [ :eachId |  "@@@@@ swap new ids with old ids?" ].
		].
	^newIsland


"
	| newIsland refId |
	newIsland := super restore: data.
	restoringFromCache = true \ifTrue: [
		refId := newIsland myName.
		newIsland send: [:i | i compactNameMap: refId.].
		self setRouterTime: (newIsland get: #time).
	].
	^newIsland."! !

!TIslandController methodsFor: 'facets' stamp: 'jrd 11/22/2008 17:19'!
revokeFacetCalled: symbolName

	| nameFacet |
	nameFacet := facets at: symbolName ifAbsent:[nil].
	nameFacet ifNotNil:[self revokeMessageFacet: nameFacet].
! !

!TIslandController methodsFor: 'facets' stamp: 'ar 9/16/2005 10:19'!
revokeMessageFacet: fid
	"Revoke the facet with the given fid"
	facets removeKey: fid ifAbsent:[].! !

!TIslandController methodsFor: 'actions' stamp: 'ar 8/31/2006 17:58'!
routerStamp: timeValue
	"ROUTERSTAMP: Change the router's time stamp"
	^self subclassResponsibility! !

!TIslandController methodsFor: 'event loop' stamp: 'jrd 1/17/2012 17:39'!
runEventLoop
	| msg |
	[true] whileTrue:[
		msg := eventQueue next.
		messageRecorder
			ifNotNilDo: [ :x | x recordMsg: msg].
		msg receiver perform: msg selector withArguments: msg arguments.
	].
! !

!TIslandController methodsFor: 'actions' stamp: 'das 11/4/2005 11:13'!
sendMessage: aTMessageData
	"SEND: Broadcast a message to all participants."
	^self subclassResponsibility! !

!TIslandController methodsFor: 'accessing' stamp: 'das 11/4/2005 11:18'!
senderID
	^senderID! !

!TIslandController methodsFor: 'accessing' stamp: 'das 11/4/2005 11:18'!
senderID: aTObjectID
	senderID := aTObjectID! !

!TIslandController methodsFor: 'accessing' stamp: 'ar 9/1/2006 14:23'!
sessionID
	"Backstop"
	^nil! !

!TIslandController methodsFor: 'initialize' stamp: 'jrd 3/12/2010 20:45'!
setupIslandSignalHandlers

	island ifNotNil: [ self runScript: #reflectReplicateValues:into: when: {island . #collectReplicateValues}. ]
! !

!TIslandController methodsFor: 'initialize' stamp: 'jrd 12/21/2009 11:00'!
shutdown

	Transcript show: '{CONTROLLER-SHUTDOWN}'.
	self shutdownIslandSignalHandlers.! !

!TIslandController methodsFor: 'initialize' stamp: 'sn 2/22/2018 18:03'!
shutdownIslandSignalHandlers

	island ifNotNil: [ "island" self stopScript: #reflectReplicateValues:into: ]! !

!TIslandController methodsFor: 'snapshot' stamp: 'sn 2/22/2018 19:49'!
snapshot
	"Create a snapshot of this island"
	| writer file |
	self critical:[
		writer := TSnapshotWriter new.
		"Prepare the island for saving"
		"island ifNil:[self newIsland]."
		island send:[:obj| obj controller: nil].
		writer snapshot: island.
		writer failed ifTrue:[^nil].
		file := (ByteArray new: 4000) writeStream.
		writer saveObjectOn: file.
		self log info: ('Snapshot Size:' , file size printString).
		island send:[:obj| obj controller: self].
	].
	^file contents! !

!TIslandController methodsFor: 'event loop' stamp: 'sn 3/1/2018 00:59'!
startEventLoop
	self stopEventLoop.
	eventLoop := self croquetFork:  #runEventLoop at: Processor userSchedulingPriority + 1.! !

!TIslandController methodsFor: 'initialize' stamp: 'mtf 1/28/2009 03:40'!
startRecieving
"Start recieving messages on this connection using the recommended settings"

	| p |
	p := self join.			"receive messages"
	p wait. "see above"
	self heartbeat: 20.
	self beServer.	"act as server"! !

!TIslandController methodsFor: 'accessing' stamp: 'ar 1/4/2006 14:58'!
status
	^status! !

!TIslandController methodsFor: 'event loop' stamp: 'ar 9/15/2005 20:51'!
stopEventLoop
	eventLoop ifNotNil:[eventLoop terminate].
	eventLoop := nil.! !

!TIslandController methodsFor: 'actions' stamp: 'ar 11/3/2005 05:56'!
sync
	"SYNC: Request a copy of the island from the router."
	^self subclassResponsibility! !

!TIslandController methodsFor: 'actions' stamp: 'ar 1/30/2006 23:39'!
sync: resID
	"SYNC: Request a copy of some resource from the router."
	^self subclassResponsibility! !

!TIslandController methodsFor: 'initialize' stamp: 'jrd 5/14/2009 14:44'!
turnLoggingOff

	log := nil! !

!TIslandController methodsFor: 'initialize' stamp: 'jrd 5/14/2009 14:44'!
turnLoggingOn

	log := true! !

!TIslandController methodsFor: 'accessing' stamp: 'jrd 1/7/2010 19:26'!
withGcDataDo: doBlock

	^gcData ifNotNil: [ doBlock value: gcData]! !

TIslandController subclass: #TLocalController
	instanceVariableNames: 'lastTick timeStamp heartbeat tickPeriod tickMsg connectedToControllers'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Objects'!

!TLocalController commentStamp: '<historical>' prior: 0!
An entirely non-replicated, loop-back kind of controller, requiring no network connectivity.!

!TLocalController methodsFor: 'actions' stamp: 'ar 11/3/2005 05:53'!
beServer
	"For local islands, this is a no-op"! !

!TLocalController methodsFor: 'initialize' stamp: 'ar 11/3/2005 06:02'!
destroy
	self stopHeartbeat.
	super destroy.! !

!TLocalController methodsFor: 'actions' stamp: 'ar 11/3/2005 06:03'!
dispatchMessage: msg
	networkQueue nextPut: msg! !

!TLocalController methodsFor: 'actions' stamp: 'das 11/3/2005 19:52'!
heartbeat: aNumber
	tickPeriod := aNumber.
	self startHeartbeat.! !

!TLocalController methodsFor: 'initialize' stamp: 'ar 1/4/2006 15:00'!
initialize
	super initialize.
	lastTick := Time millisecondClockValue.
	timeStamp := (Time totalSeconds * 1000.0) + (Time millisecondClockValue \\ 1000).
	tickPeriod := 1000. "1/sec by default"
	status := #ready. "ready by default since I'm local"! !

!TLocalController methodsFor: 'actions' stamp: 'mtf 7/23/2010 14:56'!
join
	"For local islands, this is a no-op but create a proper (e.g., resolved) promise"
	| promise |
	promise := TPromise new.
	promise resolve: true.
	^promise! !

!TLocalController methodsFor: 'actions' stamp: 'ar 11/3/2005 05:54'!
leave
	"For local islands, this is a no-op"! !

!TLocalController methodsFor: 'initialize' stamp: 'mtf 1/28/2009 03:39'!
reconnect
"For local islands, this is a no-op"! !

!TLocalController methodsFor: 'actions' stamp: 'mtf 7/23/2010 14:56'!
routerStamp: timeValue
	"ROUTERSTAMP: Change the router's time stamp"
	| p |
	timeStamp := timeValue.
	"Create a resolved promise to adhere to the protocol"
	p := TPromise new.
	p resolve: true.
	^p! !

!TLocalController methodsFor: 'heartbeat' stamp: 'ar 9/25/2006 15:21'!
runHeartbeat
	"Run the heartbeat process"
	| nextTick delta delay |
	tickMsg := TMessageData new.
	delay := Delay forMilliseconds: tickPeriod.
	[true] whileTrue:[
		"See if we've exceeded the period between ticks"
		nextTick := Time millisecondClockValue.
		delta := nextTick - lastTick.
		delta < 0 ifTrue:[delta := tickPeriod]. "clock overflow, send tick now"
		delta >= tickPeriod ifTrue:[
			self stampMessage: tickMsg.
			eventQueue nextPut:
				(MessageSend receiver: self selector: #advanceTo: argument: tickMsg time).
			delta := 0.
		].
		delay setDelayDuration: tickPeriod - delta.
		delay wait.
	].! !

!TLocalController methodsFor: 'actions' stamp: 'das 11/4/2005 11:14'!
sendMessage: msg
	"Send the message, e.g., put it into the network queue"
	self stampMessage: msg.
	self dispatchMessage: msg.! !

!TLocalController methodsFor: 'actions' stamp: 'ar 11/3/2005 12:11'!
stampMessage: msg
	"Put a time stamp on the given message"
	| nextTick delta |
	nextTick := Time millisecondClockValue.
	nextTick = lastTick ifFalse:[
		delta := nextTick - lastTick.
		"check for milliseconds clock overflow"
		delta < 0 ifTrue:[delta := 0].
		lastTick := nextTick.
		"update the current time stamp based on the number of msecs elapsed"
		timeStamp := timeStamp + delta.
	].
	msg time: timeStamp.! !

!TLocalController methodsFor: 'heartbeat' stamp: 'sn 3/1/2018 00:59'!
startHeartbeat
	"Start the heart beat process"
	self stopHeartbeat.
	heartbeat := self croquetFork: #runHeartbeat at: Processor userSchedulingPriority + 1.! !

!TLocalController methodsFor: 'heartbeat' stamp: 'ar 11/3/2005 06:00'!
stopHeartbeat
	"Start the heart beat process"
	heartbeat ifNotNil:[heartbeat terminate].
	heartbeat := nil.
! !

!TLocalController methodsFor: 'actions' stamp: 'mtf 7/23/2010 14:56'!
sync
	"For local islands, this is a no-op but create a proper (e.g., resolved) promise"
	| promise |
	promise := TPromise new.
	promise resolve: island.
	^promise! !

!TLocalController methodsFor: 'actions' stamp: 'mtf 7/23/2010 14:56'!
sync: resID
	"For local islands, this is a no-op but create a proper (e.g., resolved) promise"
	| promise |
	promise := TPromise new.
	promise resolve: nil.
	^promise! !

!TLocalController methodsFor: 'accessing' stamp: 'ar 11/3/2005 06:02'!
tickPeriod
	^tickPeriod! !

!TLocalController methodsFor: 'accessing' stamp: 'ar 11/3/2005 06:02'!
tickPeriod: msecs
	tickPeriod := msecs! !

!TLocalController methodsFor: 'accessing' stamp: 'ar 11/3/2005 06:02'!
timeStamp
	^timeStamp! !

!TLocalController methodsFor: 'accessing' stamp: 'ar 11/3/2005 06:02'!
timeStamp: time
	timeStamp := time! !

!TObject methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:32'!
extraProperties
	"This will always create the properties if they do not exist"
	^myProperties ifNil:[myProperties := self newExtraProperties].! !

!TObject methodsFor: 'private' stamp: 'ar 3/23/2006 19:32'!
myEventMap
	^myProperties ifNotNil:[myProperties eventMap]! !

!TObject methodsFor: 'private' stamp: 'ar 3/23/2006 19:32'!
myEventMap: aMap
	self extraProperties eventMap: aMap.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:39'!
myObjectAssembled
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectAssembled.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 12:00'!
myObjectAssembled:anObjectProperty
	

	|  |
	self objectProperties at: #objectAssembled put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:34'!
myObjectAttractionRepulsion
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectAttractionRepulsion.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:50'!
myObjectAttractionRepulsion:anObjectProperty
	

	|  |
	self objectProperties at: #objectAttractionRepulsion put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 10:02'!
myObjectCollisions
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectCollisions.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 10:15'!
myObjectCollisions:oc
	"comment stating purpose of message"

	|  |
	self objectProperties at: #objectCollisions put: oc..! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:39'!
myObjectComponents
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectComponents.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 12:01'!
myObjectComponents:anObjectProperty
	

	|  |
	self objectProperties at: #objectComponents put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:38'!
myObjectCreator
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectCreator.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:59'!
myObjectCreator:anObjectProperty
	

	|  |
	self objectProperties at: #objectCreator put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:35'!
myObjectElasticity
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectElasticity.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:51'!
myObjectElasticity:anObjectProperty
	

	|  |
	self objectProperties at: #objectElasticity put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:32'!
myObjectInteractivityClass
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectInteractivityClass.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:49'!
myObjectInteractivityClass:anObjectProperty
	

	|  |
	self objectProperties at: #objectInteractivityClass put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:38'!
myObjectKeyWords
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectKeyWords.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:57'!
myObjectKeyWords:anObjectProperty
	

	|  |
	self objectProperties at: #objectKeyWords put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:24'!
myObjectLoopSound
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectLoopSound.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:46'!
myObjectLoopSound:anObjectProperty
	

	|  |
	self objectProperties at: #objectLoopSound put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:23'!
myObjectMakeGroupable
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectMakeGroupable.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:44'!
myObjectMakeGroupable:anObjectProperty
	"comment stating purpose of message"

	|  |
	self objectProperties at: #objectMakeGroupable put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:22'!
myObjectMakeMovable
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectMakeMovable.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:43'!
myObjectMakeMovable:anObjectProperty
	"comment stating purpose of message"

	|  |
	self objectProperties at: #objectMakeMovable put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:34'!
myObjectMass
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectMass.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:50'!
myObjectMass:anObjectProperty
	

	|  |
	self objectProperties at: #objectMass put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:31'!
myObjectMotion
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectMotion.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:48'!
myObjectMotion:anObjectProperty
	

	|  |
	self objectProperties at: #objectMotion put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:39'!
myObjectPermissions
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectPermissions.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 12:01'!
myObjectPermissions:anObjectProperty
	

	|  |
	self objectProperties at: #objectPermissions put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:32'!
myObjectPhysics
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectPhysics.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:48'!
myObjectPhysics:anObjectProperty
	

	|  |
	self objectProperties at: #objectPhysics put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:23'!
myObjectSound
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectSound.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:45'!
myObjectSound:anObjectProperty
	

	|  |
	self objectProperties at: #objectSound put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:30'!
myObjectSoundOnClick
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectSoundOnClick.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:47'!
myObjectSoundOnClick:anObjectProperty
	

	|  |
	self objectProperties at: #objectSoundOnClick put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:31'!
myObjectSoundOnCollision
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectSoundOnCollision.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:47'!
myObjectSoundOnCollision:anObjectProperty
	

	|  |
	self objectProperties at: #objectSoundOnCollision put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 12:28'!
myObjectSoundVolume
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectSoundVolume.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:45'!
myObjectSoundVolume:anObjectProperty
	

	|  |
	self objectProperties at: #objectSoundVolume put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:31'!
myObjectSpeed
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectSpeed.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:48'!
myObjectSpeed:anObjectProperty
	

	|  |
	self objectProperties at: #objectSpeed put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:34'!
myObjectSurfaceFriction
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectSurfaceFriction.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:50'!
myObjectSurfaceFriction:anObjectProperty
	

	|  |
	self objectProperties at: #objectSurfaceFriction put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:40'!
myObjectTexture
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectTexture.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:43'!
myObjectTexture:anObjectProperty
	"comment stating purpose of message"

	|  |
	self objectProperties at: #objectTexture put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:37'!
myObjectThingDescription
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectThingDescription.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:57'!
myObjectThingDescription:anObjectProperty
	

	|  |
	self objectProperties at: #objectThingDescription put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:37'!
myObjectThingName
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectThingName.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:56'!
myObjectThingName:anObjectProperty
	

	|  |
	self objectProperties at: #objectThingName put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:38'!
myObjectType
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectType.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:58'!
myObjectType:anObjectProperty
	

	|  |
	self objectProperties at: #objectType put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:39'!
myObjectURL
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectURL.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 12:01'!
myObjectURL:anObjectProperty
	

	|  |
	self objectProperties at: #objectURL put: anObjectProperty.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:35'!
myObjectWindFriction
	"comment stating purpose of message"

	|  |
	^self objectProperties at: #objectWindFriction.! !

!TObject methodsFor: 'properties dictionary access' stamp: 'eb 2/5/2009 11:51'!
myObjectWindFriction:anObjectProperty
	

	|  |
	self objectProperties at: #objectWindFriction put: anObjectProperty.! !

!TObject methodsFor: 'private' stamp: 'ar 3/23/2006 19:33'!
myScripts
	^(myProperties ifNotNil:[myProperties scripts]) ifNil:[#()]! !

!TObject methodsFor: 'private' stamp: 'ar 3/23/2006 19:33'!
myScripts: scripts
	self extraProperties scripts: scripts.! !

!TObject methodsFor: 'private' stamp: 'eb 2/5/2009 09:52'!
newCobaltObjectProperties
	"Create new extra properties for this object"
	^CobaltObjectPropertiesDictionary new.! !

!TObject methodsFor: 'private' stamp: 'ar 3/23/2006 19:34'!
newExtraProperties
	"Create new extra properties for this object"
	^TObjectProperties new.! !

!TObject methodsFor: 'accessing' stamp: 'eb 2/5/2009 09:54'!
objectProperties
	"This will always create the properties if they do not exist"
	^myCobaltObjectProperties ifNil:[myCobaltObjectProperties := self newCobaltObjectProperties].! !

Object subclass: #TObjectProperties
	instanceVariableNames: 'eventMap scripts properties'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TObjectProperties commentStamp: '<historical>' prior: 0!
I represent some optional state for TObjects which is allocated lazily if needed.!

!TObjectProperties methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:33'!
eventMap
	^eventMap! !

!TObjectProperties methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:33'!
eventMap: aMap
	eventMap := aMap.! !

!TObjectProperties methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:33'!
scripts
	^scripts! !

!TObjectProperties methodsFor: 'accessing' stamp: 'ar 3/23/2006 19:34'!
scripts: anArray
	scripts := anArray.! !

Object subclass: #TPromise
	instanceVariableNames: 'myIsland myName myScripts resolved broken result resolver resolvers handlers'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Objects'!

!TPromise commentStamp: '<historical>' prior: 0!
I represent a promise being used when a client needs the result of a future message. In that situation the client can use the promise to execute code when the message is actually executed.
Note that this implementation assumes one consumer of the promise.  If you wait twice on the promise one of the consumers will wait forever. Similarly for whenComplete: - i you try to add 2 when complete blocks, only the last one added will work.

Instance variables:
	myIsland	<>	
	myName	<>
	myScripts	<Array>		Cached script registry for delayed execution
	resolved	<Boolean>	True if the message has been resolved.
	broken		<Boolean>	True if the message has been broken.
	result		<Object>		The result of the promise (often a FarRef), or the exception, if broken
	resolver	<Block>		The code to be executed when the promise completes.
	resolvers	<Array of Block>		The code to be executed when the promise completes.
	handlers	<Array of Block>		The code to be executed when the promise completes.

Example:
	island := TIsland new.					"creates a new island"
	promise := island future new: Point.		"creates a promise representing the #new: message"
	promise whenComplete:[:result|			"code being executed when promise completed"
		result future x: 1.
		result future y: 2.
	].

!

!TPromise methodsFor: 'accessing' stamp: 'ar 11/8/2005 17:01'!
id
	^myName! !

!TPromise methodsFor: 'accessing' stamp: 'ar 11/8/2005 17:01'!
id: aTObjectID
	myName := aTObjectID! !

!TPromise methodsFor: 'testing' stamp: 'ar 9/16/2005 13:50'!
isResolved
	^resolved == true! !

!TPromise methodsFor: 'accessing' stamp: 'ar 11/8/2005 16:56'!
island
	^myIsland! !

!TPromise methodsFor: 'accessing' stamp: 'ar 11/8/2005 16:56'!
island: anIsland
	myIsland := anIsland! !

!TPromise methodsFor: 'accessing' stamp: 'ar 9/8/2005 12:06'!
myScripts
	^myScripts ifNil:[#()]! !

!TPromise methodsFor: 'accessing' stamp: 'ar 9/8/2005 12:05'!
myScripts: anArray
	myScripts := anArray! !

!TPromise methodsFor: 'resolving' stamp: 'NiccoKunzmann 6/2/2014 17:13'!
onResolved: aBoolean
	"This could possibly be to block? I do not know"
	"Let us try this:"
	self assert: self isResolved not.
	self resolve: aBoolean
	"
CobaltController>>recvFacets:from:facet: 
shows:
loginPromise resolve: true."! !

!TPromise methodsFor: 'initialize' stamp: 'ar 11/8/2005 17:01'!
register
	"register me"
	Processor activeIsland at: myName put: self. "Eeek!!"
! !

!TPromise methodsFor: 'resolving' stamp: 'mtf 7/23/2010 14:55'!
resolve: arg
	"Resolve this promise"
	self unregister.
	result := arg.
	resolver ifNotNil:[
		resolver numArgs = 0 ifTrue:[resolver value].
		resolver numArgs = 1 ifTrue:[resolver value: arg].
		resolver := nil.
	].
	resolved := true.! !

!TPromise methodsFor: 'accessing' stamp: 'ar 9/14/2005 17:59'!
result
	^result! !

!TPromise methodsFor: 'initialize' stamp: 'mtf 7/23/2010 14:57'!
unregister
	Processor activeIsland globals removeKey: myName ifAbsent:[]. "eeek!!!!!!"
	self pauseScript: #resolve:.! !

!TPromise methodsFor: 'accessing' stamp: 'ar 11/8/2005 18:06'!
value
	^result! !

!TPromise methodsFor: 'waiting' stamp: 'das 1/6/2006 15:12'!
wait
	"Wait unconditionally for this promise to resolve."
	| sema |
	sema := Semaphore new.
	self whenComplete:[sema signal].
	sema wait.! !

!TPromise methodsFor: 'waiting' stamp: 'das 1/6/2006 15:12'!
waitTimeoutMSecs: msecs
	"Wait for at most the given number of milliseconds for this promise to resolve. Answer true if it is resolved, false otherwise."
	| sema |
	sema := Semaphore new.
	self whenComplete:[sema signal].
	Delay timeoutSemaphore: sema afterMSecs: msecs.
	^self isResolved! !

!TPromise methodsFor: 'resolving' stamp: 'mtf 7/23/2010 14:57'!
whenComplete: aBlock
	"Evaluate aBlock when I am resolved"
	aBlock numArgs <= 2 ifFalse:[self error: 'Must be less than three argument block'].
	resolver := aBlock.
	self isResolved ifTrue:[self resolve: result].! !

Object subclass: #TResourceManager
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

TResourceManager class
	instanceVariableNames: 'default'!

!TResourceManager commentStamp: '<historical>' prior: 0!
Placeholder class for future expansion of the cache/management subsystems. Currently it is kinda weak and we need something more structured along the lines of the OGRE resourece managers!

TResourceManager class
	instanceVariableNames: 'default'!

TResourceManager subclass: #TFileCacheManager
	instanceVariableNames: 'mutex location'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Kernel-Support'!

!TFileCacheManager commentStamp: '<historical>' prior: 0!
A simple cache manager looking in a particular directory for some cached resource.

Instance variables:
	mutex	<TMutex>	Mutex for concurrent access.
	location	<String>		Relative uri (using forward slashes) for the cache location.!

!TFileCacheManager methodsFor: 'searching' stamp: 'ar 1/30/2006 12:37'!
cache: data to: fileName
	"Find the file with the given resource id"
	| file |
	mutex critical:[
		file := [self directory forceNewFileNamed: fileName] on: Error do:[:ex| ex return: nil].
		file ifNotNil:[
			[file binary.
			file nextPutAll: data] ensure:[file close].
		].
	].
	^data! !

!TFileCacheManager methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:39'!
directory
	| dir |
	dir := FileDirectory on: (FileDirectory default pathFromURI: location).
	dir assureExistence.
	^dir! !

!TFileCacheManager methodsFor: 'searching' stamp: 'ar 1/30/2006 12:29'!
find: fileName
	"Find the file with the given resource id"
	| file data |
	mutex critical:[
		file := [self directory readOnlyFileNamed: fileName] on: Error do:[:ex| ex return: nil].
		file ifNotNil:[
			[file binary.
			data := file contents] ensure:[file close].
		].
	].
	^data! !

!TFileCacheManager methodsFor: 'initialize' stamp: 'ar 1/30/2006 12:48'!
initialize
	super initialize.
	mutex := TMutex new.
	location := 'cache'! !

!TFileCacheManager methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:24'!
location
	^location! !

!TFileCacheManager methodsFor: 'accessing' stamp: 'ar 1/30/2006 12:24'!
location: aString
	location := aString! !

!TFileCacheManager methodsFor: 'searching' stamp: 'ar 1/30/2006 13:09'!
readOnlyFileNamed: fileName
	"Find the file with the given resource id"
	| file |
	mutex critical:[
		file := [self directory readOnlyFileNamed: fileName] on: Error do:[:ex| ex return: nil].
	].
	^file! !

!TResourceManager class methodsFor: 'as yet unclassified' stamp: 'jrd 9/22/2010 20:12'!
defaultFormPath
	
	^FileDirectory fullPathForURI: 'resources/textures/', 'default.bmp'! !
TIsland initialize!
TCryptoRandom initialize!
TFarRef initialize!
TLogger initialize!
TIslandController initialize!
